"""
=======================================
Solution (:mod:`CADETProcess.solution`)
=======================================

.. currentmodule:: CADETProcess.solution

Module to store and plot solution of simulation.

.. autosummary::
   :toctree: generated/

   SolutionBase
   SolutionIO
   SolutionBulk
   SolutionParticle
   SolutionSolid
   SolutionVolume


Method to slice a Solution:

.. autosummary::
   :toctree: generated/

   slice_solution
   slice_solution_front


"""  # noqa

from __future__ import annotations

import copy
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.collections import QuadMesh
from scipy import integrate
from scipy.interpolate import PchipInterpolator

from CADETProcess import CADETProcessError, plotting, smoothing, transform
from CADETProcess.dataStructure import (
    SizedNdArray,
    String,
    Structure,
    Typed,
    UnsignedFloat,
    Vector,
)
from CADETProcess.dynamicEvents import TimeLine
from CADETProcess.processModel import ComponentSystem

__all__ = [
    "SolutionIO",
    "SolutionBulk",
    "SolutionParticle",
    "SolutionSolid",
    "SolutionVolume",
    "slice_solution",
    "slice_solution_front",
]


class SolutionBase(Structure):
    """
    Base class for solutions of component systems.

    This class represents a solution of a component system at different time points.
    It provides several properties that allow access to information about the solution,
    such as the number of components, the total concentration, and the local purity.

    Attributes
    ----------
    name : str
        Name of the solution.
    time : np.ndarray
        Array of time points.
    solution : np.ndarray
        Array of solution values.
    c_min : float
        Minimum concentration threshold, below which concentrations are considered zero.
    dimensions : list of str
        Names of the dimensions in the solution.

    Notes
    -----
    This class is not meant to be used directly, but to be subclassed by specific
    solution types.
    """

    name = String()
    time = Vector()
    component_system = Typed(ty=ComponentSystem)
    solution = SizedNdArray(size="solution_shape")
    c_min = UnsignedFloat(default=1e-6)

    dimensions = ["time"]

    def __init__(
        self,
        name: str,
        component_system: ComponentSystem,
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
    ) -> None:
        """
        Initialize the SolutionBase.

        Parameters
        ----------
        name : str
            The name of the solution.
        component_system : ComponentSystem
            The component system associated with the solution.
        time : npt.ArrayLike
            An array-like structure representing time points.
        solution : npt.ArrayLike
            An array-like structure representing the solution data.
        """
        self.name = name
        self.component_system = component_system
        self.time = time
        self.solution = solution

        self.update_solution()

    def update_solution(self) -> None:
        """Update the solution."""
        pass

    def update_transform(self) -> None:
        """Update the transforms."""
        pass

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def cycle_time(self) -> float:
        """float: Cycle time."""
        return self.time[-1]

    @property
    def nt(self) -> int:
        """int: Number of time steps."""
        return len(self.time)

    @property
    def component_coordinates(self) -> np.ndarray:
        """np.ndarray: Indices of the components."""
        return np.arange(self.n_comp)

    @property
    def coordinates(self) -> dict[str, np.ndarray]:
        """dict[str, np.ndarray]: Coordinates of the Solution."""
        coordinates = {}

        for c in self.dimensions:
            v = getattr(self, c)
            if v is None:
                continue
            coordinates[c] = v
        return coordinates

    @property
    def solution_shape(self) -> tuple[int]:
        """tuple[int]: (Expected) shape of the solution."""
        return tuple(len(c) for c in self.coordinates.values())

    @property
    def total_concentration_components(self) -> np.ndarray:
        """np.ndarray: Total concentration of components (sum of species)."""
        component_concentration = np.zeros(
            self.solution.shape[0:-1] + (self.component_system.n_components,)
        )

        counter = 0
        for index, comp in enumerate(self.component_system):
            comp_indices = slice(counter, counter + comp.n_species)
            c_comp = np.sum(self.solution[..., comp_indices], axis=-1)
            component_concentration[..., index] = c_comp
            counter += comp.n_species

        return component_concentration

    @property
    def total_concentration(self) -> np.ndarray:
        """np.ndarray: Total concentration (sum) of all components."""
        return np.sum(self.solution, axis=-1, keepdims=True)

    @property
    def local_purity_components(self) -> np.ndarray:
        """np.ndarray: Local purity of each component."""
        solution = self.total_concentration_components
        c_total = self.total_concentration
        c_total[c_total < self.c_min] = np.nan

        purity = np.zeros(solution.shape)
        with np.errstate(divide="ignore", invalid="ignore"):
            purity = solution / c_total

        purity = np.nan_to_num(purity)

        return purity

    @property
    def local_purity_species(self) -> np.ndarray:
        """np.ndarray: Local purity of components."""
        solution = self.solution
        c_total = self.total_concentration
        c_total[c_total < self.c_min] = np.nan

        purity = np.zeros(solution.shape)
        with np.errstate(divide="ignore", invalid="ignore"):
            purity = solution / c_total

        purity = np.nan_to_num(purity)

        return purity

    def __str__(self) -> str:
        return self.name


class SolutionIO(SolutionBase):
    """
    Solution representing streams at the inlet or outlet of a ``UnitOperation``.

    Notes
    -----
    The `flow_rate` attribute is implemented as TimeLine to improve interpolation of
    signals with discontinuous flow.
    """

    dimensions = SolutionBase.dimensions + ["component_coordinates"]

    def __init__(
        self,
        name: str,
        component_system: ComponentSystem,
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
        flow_rate: TimeLine | npt.ArrayLike,
    ) -> None:
        """
        Initialize the SolutionIO.

        Parameters
        ----------
        name : str
            The name of the solution object.
        component_system : ComponentSystem
            The component system associated with the solution.
        time : npt.ArrayLike
            An array-like structure representing time points.
        solution : npt.ArrayLike
            An array-like structure representing the solution data.
        flow_rate : TimeLine | npt.ArrayLike
            The flow rate data, which can be a TimeLine object or an array-like structure.
        """
        if not isinstance(flow_rate, TimeLine):
            flow_rate = TimeLine.from_profile(time, flow_rate)
        self.flow_rate = flow_rate

        super().__init__(name, component_system, time, solution)

    def update_solution(self) -> None:
        """Update solution method."""
        self._solution_interpolated = None
        self._dm_dt_interpolated = None
        self.update_transform()

    @property
    def derivative(self) -> "SolutionIO":
        """SolutionIO: Derivative of this solution."""
        derivative = copy.deepcopy(self)
        derivative_fun = derivative.solution_interpolated.derivative
        derivative.solution = derivative_fun(derivative.time)
        derivative.update_solution()

        return derivative

    @property
    def antiderivative(self) -> "SolutionIO":
        """SolutionIO: Antiderivative of this solution."""
        antiderivative = copy.deepcopy(self)
        antiderivative_fun = antiderivative.solution_interpolated.antiderivative
        antiderivative.solution = antiderivative_fun(antiderivative.time)
        antiderivative.update_solution()

        return antiderivative

    def update_transform(self) -> None:
        """Update the Transformer."""
        self.transform = transform.NormLinearTransformer(
            np.min(self.solution, axis=0),
            np.max(self.solution, axis=0),
            allow_extended_input=True,
            allow_extended_output=True,
        )

    @property
    def solution_interpolated(self) -> InterpolatedSignal:
        """InterpolatedSignal: The interpolated signal."""
        if self._solution_interpolated is None:
            self._solution_interpolated = InterpolatedSignal(self.time, self.solution)

        return self._solution_interpolated

    @property
    def dm_dt_interpolated(self) -> InterpolatedSignal:
        """Return the dm/dt interpolated signal."""
        if self._dm_dt_interpolated is None:
            dm_dt = self.solution * self.flow_rate.value(self.time)
            self._dm_dt_interpolated = InterpolatedSignal(self.time, dm_dt)

        return self._dm_dt_interpolated

    def normalize(self) -> SolutionIO:
        """SolutionIO: Normalize the solution using the transformation function."""
        solution = copy.deepcopy(self)

        solution.solution = self.transform.transform(self.solution)
        solution.update_solution()

        return solution

    def resample(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        nt: Optional[int] = 5001,
    ) -> "SolutionIO":
        """
        Resample solution to nt time points.

        Parameters
        ----------
        start : Optional[float]
            Start time of the fraction. If None, the first time point is used. The
            default is None.
        end: Optional[float]
            End time of the fraction. If None, the last time point is used. The
            default is None.
        nt : int, optional
            Number of points to resample. The default is 5001.

        Returns
        -------
        SolutionIO
            The resampled solution object.
        """
        solution = copy.deepcopy(self)

        if start is None:
            start = self.time[0]
        if end is None:
            end = self.time[-1]

        solution.time = np.linspace(start, end, nt)
        solution.solution = self.solution_interpolated(solution.time)
        solution.update_solution()

        return solution

    def smooth_data(
        self,
        s: Optional[float | list[float]] = None,
        crit_fs: Optional[float | list[float]] = None,
        crit_fs_der: Optional[float | list[float]] = None,
    ) -> "SolutionIO":
        """
        Smooth data.

        Parameters
        ----------
        s : Optional[float | list[float]]
            DESCRIPTION.
        crit_fs : Optional[float | list[float]]
            DESCRIPTION.
        crit_fs_der : Optional[float | list[float]]
            DESCRIPTION.

        Returns
        -------
        SolutionIO
            The smoothed solution object.
        """
        solution = copy.deepcopy(self)

        if None in (s, crit_fs, crit_fs_der):
            s_ = []
            crit_fs_ = []
            crit_fs_der_ = []

            for i in range(self.n_comp):
                s_i, crit_fs_i, crit_fs_der_i = smoothing.find_smoothing_factors(
                    self.time, self.solution[..., i], rmse_target=1e-3
                )
                s_.append(s_i)
                crit_fs_.append(crit_fs_i)
                crit_fs_der_.append(crit_fs_der_i)

            if s is None:
                s = s_
            if crit_fs is None:
                crit_fs = crit_fs_
            if crit_fs_der is None:
                crit_fs_der = crit_fs_der_

        if np.isscalar(s):
            s = self.n_comp * [s]
        elif len(s) == 1:
            s = self.n_comp * s

        if np.isscalar(crit_fs):
            crit_fs = self.n_comp * [crit_fs]
        elif len(crit_fs) == 1:
            crit_fs = self.n_comp * crit_fs

        if np.isscalar(crit_fs_der):
            crit_fs_der = self.n_comp * [crit_fs_der]
        elif len(crit_fs_der) == 1:
            crit_fs_der = self.n_comp * crit_fs_der

        solution = np.zeros((self.solution.shape))
        for i, (s, crit_fs, crit_fs_der) in enumerate(zip(s, crit_fs, crit_fs_der)):
            smooth, smooth_der = smoothing.full_smooth(
                self.time, self.solution[..., i], crit_fs, s, crit_fs_der
            )
            solution[..., i] = smooth

        solution.solution = solution
        solution.update_solution()

        return solution

    def integral(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
    ) -> np.ndarray:
        """
        Peak area in a fraction interval.

        Parameters
        ----------
        start : Optional[float]
            Start time of the fraction. If None, the first time point is used. The
            default is None.
        end: Optional[float]
            End time of the fraction. If None, the last time point is used. The
            default is None.

        Returns
        -------
        np.ndarray
            Area of each component in the fraction.
        """
        if start is None:
            start = self.time[0]

        if end is None:
            end = self.cycle_time

        return self.solution_interpolated.integral(start, end)

    def create_fraction(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
    ) -> Fraction:  # noqa:F821
        """
        Create fraction in interval [start, end].

        Parameters
        ----------
        start : Optional[float]
            Start time of the fraction. If None, the first time point is used. The
            default is None.
        end: Optional[float]
            End time of the fraction. If None, the last time point is used. The
            default is None.

        Returns
        -------
        Fraction
            a Fraction object in the interval [start, end]
        """
        if start is None:
            start = self.time[0]

        if end is None:
            end = self.cycle_time

        from CADETProcess.fractionation import Fraction

        mass = self.fraction_mass(start, end)
        volume = self.fraction_volume(start, end)
        return Fraction(mass, volume, start, end)

    def fraction_mass(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
    ) -> np.ndarray:
        """
        Component mass in a fraction interval.

        Parameters
        ----------
        start : Optional[float]
            Start time of the fraction. If None, the first time point is used. The
            default is None.
        end: Optional[float]
            End time of the fraction. If None, the last time point is used. The
            default is None.

        Returns
        -------
        np.ndarray
            Mass of all components in the fraction.
        """
        if start is None:
            start = self.time[0]

        if end is None:
            end = self.cycle_time

        # Note, we do not use self.dm_dt_interpolated to better account for
        # discontinuities in the flow rate profile, by passing the section times to
        # `quad_vec`. Maybe this can be improved in the future.
        def dm_dt(t: float, flow_rate: np.ndarray, solution: np.ndarray) -> np.ndarray:
            dm_dt = flow_rate.value(t) * solution(t)
            return dm_dt

        points = None
        if len(self.flow_rate.section_times) > 2:
            points = self.flow_rate.section_times[1:-1]

        mass = integrate.quad_vec(
            dm_dt,
            start,
            end,
            epsabs=1e-6,
            epsrel=1e-8,
            args=(self.flow_rate, self.solution_interpolated),
            points=points,
        )[0]

        return mass

    def fraction_volume(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
    ) -> float:
        """
        Volume of a fraction interval.

        Parameters
        ----------
        start : Optional[float]
            Start time of the fraction. If None, the first time point is used. The
            default is None.
        end: Optional[float]
            End time of the fraction. If None, the last time point is used. The
            default is None.

        Returns
        -------
        float
            Volume of the fraction
        """
        if start is None:
            start = self.time[0]

        if end is None:
            end = self.cycle_time

        return float(self.flow_rate.integral(start, end).squeeze())

    @plotting.create_and_save_figure
    def plot(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        y_max: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the entire time_signal for each component.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
            The default is None.
        layout : Optional[plotting.Layout]
            Plot layout options. The default is None.
        y_max : Optional[float]
            Maximum value of the y-axis. If None, the value is automatically
            determined from the data. The default is None.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes.
        ax : Optional[Axes]
            Axes to plot on. If None, a new figure is created.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plotlib
        plot_purity
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [start, end]},
        )

        x = solution.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$c~/~mM$"
            layout.x_lim = (start, end)
        if y_max is not None:
            layout.y_lim = (None, y_max)

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        return ax

    @plotting.create_and_save_figure
    def plot_purity(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        y_max: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        plot_components_purity: Optional[bool] = True,
        plot_species_purity: Optional[bool] = False,
        alpha: Optional[float] = 1,
        hide_labels: Optional[bool] = False,
        show_legend: Optional[bool] = True,
        ax: Optional[Axes] = None,
    ) -> Axes:
        """
        Plot local purity for each component of the concentration profile.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
            Note that if components are excluded, they will also not be considered in
            the calculation of the purity.
        layout : Optional[plotting.Layout]
            Plot layout options.
        y_max : Optional[float]
            Maximum value of y axis.
            If None, value is automatically deferred from solution.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes.
        plot_components_purity : Optional[bool], default=True
            If True, plot purity of total component concentration.
        plot_species_purity : Optional[bool], default=False
            If True, plot purity of individual species.
        alpha : Optional[float], default=1
            Opacity of line.
        hide_labels : Optional[bool], default=False
            If True, hide labels.
        show_legend : Optional[bool], default=True
            If True, show legend.
        ax : Optional[Axes]
            Axes to plot on.

        Returns
        -------
        Axes
            Axes with plot of purity over time.

        Raises
        ------
        CADETProcessError
            If solution has less than 2 components.

        See Also
        --------
        slice_solution
        plotlib
        plot
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [start, end]},
        )

        if solution.n_comp < 2:
            raise CADETProcessError(
                "Purity undefined for systems with less than 2 components."
            )

        x = solution.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = r"$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = r"$time~/~min$"
            layout.y_label = r"$Purity ~/~\%$"
            if start is not None:
                start /= 60
            if end is not None:
                end /= 60
            layout.x_lim = (start, end)
        if y_max is not None:
            layout.y_lim = (None, y_max)

        local_purity_components = solution.local_purity_components * 100
        local_purity_species = solution.local_purity_species * 100

        colors = iter(plt.rcParams["axes.prop_cycle"].by_key()["color"])
        species_index = 0
        for i, comp in enumerate(solution.component_system.components):
            color = next(colors)
            if hide_labels:
                label = None
            else:
                label = comp.name

            if plot_components_purity:
                y = local_purity_components[..., i]

                ax.plot(
                    x,
                    y,
                    label=label,
                    color=color,
                    alpha=alpha,
                )

            if plot_species_purity:
                if comp.n_species == 1:
                    species_index += 1
                    continue

                for s, species in enumerate(comp.species):
                    label = s

                    y = local_purity_species[..., species_index]

                    ax.plot(
                        x,
                        y,
                        "--",
                        label=label,
                        color=color,
                        alpha=alpha,
                    )
                    species_index += 1

        plotting.set_layout(
            ax,
            layout,
            show_legend,
        )

        return ax


class SolutionBulk(SolutionBase):
    """
    Solution in the bulk phase of the ``UnitOperation``.

    Dimensions: NCOL * NRAD * NCOMP
    """

    dimensions = SolutionBase.dimensions + [
        "axial_coordinates",
        "radial_coordinates",
        "component_coordinates",
    ]

    def __init__(
        self,
        name: str,
        component_system: ComponentSystem,
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
        axial_coordinates: Optional[npt.ArrayLike] = None,
        radial_coordinates: Optional[npt.ArrayLike] = None,
    ) -> None:
        """
        Initialize the SolutionBulk.

        Parameters
        ----------
        name : str
            The name of the solution object.
        component_system : ComponentSystem
            The component system associated with the solution.
        time : npt.ArrayLike
            An array-like structure representing time points.
        solution : npt.ArrayLike
            An array-like structure representing the solution data.
        axial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing axial coordinates.
            If None, it is assumed that the model has a singleton axial dimension.
        radial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing radial coordinates.
            If None, it is assumed that the model has a singleton bulk dimension.
        """
        self.axial_coordinates = axial_coordinates
        self.radial_coordinates = radial_coordinates

        super().__init__(name, component_system, time, solution)

    @property
    def ncol(self) -> int:
        """int: Number of axial discretization points."""
        if self.axial_coordinates is None:
            return
        return len(self.axial_coordinates)

    @property
    def nrad(self) -> int:
        """int: Number of radial discretization points."""
        if self.radial_coordinates is None:
            return
        return len(self.radial_coordinates)

    @plotting.create_and_save_figure
    def plot(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        y_max: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the entire time_signal for each component.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
        y_max : Optional[float]
            Maximum value of y axis.
            If None, value is automatically deferred from solution.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        Raises
        ------
        CADETProcessError
            If solution is not 1D.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plotlib
        plot_purity
        """
        if not (self.ncol is None and self.nrad is None):
            raise CADETProcessError(
                "Solution has more than single dimension. "
                "Please use `plot_at_time` or `plot_at_position`."
            )

        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [start, end]},
        )

        x = solution.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$c~/~mM$"
            layout.x_lim = (start, end)
            if y_max is not None:
                layout.y_lim = (None, y_max)

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        return ax

    @plotting.create_and_save_figure
    def plot_at_time(
        self,
        t: float,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot bulk solution over space at given time.

        Parameters
        ----------
        t : float
            Time for plotting in seconds.
            If t == -1, the final solution is plotted.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plot_at_position
        plotlib
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [t, t]},
        )

        x = self.axial_coordinates

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$z~/~m$"
            layout.y_label = "$c~/~mM$"

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        plotting.add_text(ax, f"time = {t:.2f} s")

        return ax

    @plotting.create_and_save_figure
    def plot_at_position(
        self,
        z: float,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        x_axis_in_minutes: bool = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot bulk solution over time at given position.

        Parameters
        ----------
        z : float
            Position for plotting.
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
            If None, value is automatically deferred from solution.
        x_axis_in_minutes : bool
            If True, the x-axis will be plotted using minutes. The default is True.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plot_at_position
        plotlib
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"axial_coordinates": [z, z]},
        )

        x = self.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        x = self.time / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$c~/~mM$"

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        plotting.add_text(ax, f"z = {z:.2f} m")

        return ax


class SolutionParticle(SolutionBase):
    """
    Solution in the particle liquid phase of the ``UnitOperation``.

    Dimensions: NCOL * NRAD * sum_{j}^{NPARTYPE}{NCOMP * NPAR,j}
    """

    dimensions = SolutionBase.dimensions + [
        "axial_coordinates",
        "radial_coordinates",
        "particle_coordinates",
        "component_coordinates",
    ]

    def __init__(
        self,
        name: str,
        component_system: ComponentSystem,
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
        axial_coordinates: Optional[npt.ArrayLike] = None,
        radial_coordinates: Optional[npt.ArrayLike] = None,
        particle_coordinates: Optional[npt.ArrayLike] = None,
    ) -> None:
        """
        Initialize the SolutionParticle.

        Parameters
        ----------
        name : str
            The name of the solution object.
        component_system : ComponentSystem
            The component system associated with the solution.
        time : npt.ArrayLike
            An array-like structure representing time points.
        solution : npt.ArrayLike
            An array-like structure representing the solution data.
        axial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing axial coordinates.
            If None, it is assumed that the model has a singleton axial dimension.
        radial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing radial coordinates.
            If None, it is assumed that the model has a singleton bulk dimension.
        particle_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing particle coordinates.
            If None, it is assumed that the model has a singleton particle dimension.
        """
        self.axial_coordinates = axial_coordinates
        self.radial_coordinates = radial_coordinates
        self.particle_coordinates = particle_coordinates

        super().__init__(name, component_system, time, solution)

    @property
    def ncol(self) -> int:
        """int: Number of axial discretization points."""
        if self.axial_coordinates is None:
            return
        return len(self.axial_coordinates)

    @property
    def nrad(self) -> int:
        """int: Number of radial discretization points."""
        if self.radial_coordinates is None:
            return
        return len(self.radial_coordinates)

    @property
    def npar(self) -> int:
        """int: Number of particle discretization points."""
        if self.particle_coordinates is None:
            return
        return len(self.particle_coordinates)

    @plotting.create_and_save_figure
    def plot(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        y_max: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the entire particle liquid phase solution for each component.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
        y_max : Optional[float]
            Maximum value of y axis.
            If None, value is automatically deferred from solution.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes. The default is True.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        ax : Axes
            Axes object with concentration profile.

        Raises CADETProcessError
            If solution is not 1D.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plotlib
        plot_purity
        """
        if not (self.ncol is None and self.nrad is None and self.npar is None):
            raise CADETProcessError(
                "Solution has more than single dimension. Please use `plot_at_time`."
            )

        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [start, end]},
        )

        x = solution.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$c~/~mM$"
            layout.x_lim = (start, end)
        if y_max is not None:
            layout.y_lim = (None, y_max)

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        return ax

    def _plot_1D(
        self,
        t: float,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot particle solution over space at given time.

        Parameters
        ----------
        t : float
            Time for plotting.
            If t == -1, the final solution is plotted.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        _plot_solution_1D
        _plot_2D
        slice_solution
        plotlib
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [t, t]},
        )

        x = self.axial_coordinates

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$z~/~m$"
            layout.y_label = "$c~/~mM$"

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        plotting.add_text(ax, f"time = {t:.2f} s")

        return ax

    def _plot_2D(
        self,
        ax: Axes,
        t: float,
        comp: int = 0,
        vmax: Optional[float] = None,
    ) -> Axes:
        x = self.axial_coordinates
        y = self.particle_coordinates

        if not self.time[0] <= t <= self.time[-1]:
            raise ValueError("Time exceeds bounds.")
        i = np.where(t <= self.time)[0][0]
        v = self.solution[i, :, :, comp].transpose()

        if vmax is None:
            vmax = v.max()
        try:
            mesh = ax.get_children()[0]
            mesh.set_array(v.flatten())
        except AttributeError:
            mesh = ax.pcolormesh(x, y, v, shading="gouraud", vmin=0, vmax=vmax)

        plotting.add_text(ax, f"time = {t:.2f} s")

        layout = plotting.Layout()
        layout.x_label = "$z~/~m$"
        layout.y_label = "$r~/~m$"
        layout.title = f"Solid phase concentration, comp={comp}"

        plotting.set_layout(ax, layout)
        plt.colorbar(mesh)

        return ax

    @plotting.create_and_save_figure
    def plot_at_time(
        self,
        t: float,
        comp: Optional[int] = None,
        vmax: Optional[float] = None,
        ax: Optional[Axes] = None,
    ) -> Axes:
        """
        Plot particle liquid solution for a given component over space at given time.

        Parameters
        ----------
        t : float
            Solution time at with to plot.
        comp : Optional[int]
            Component index.
        vmax: Optional[float]
            Maximum data value to scale color map.
        ax : Optional[Axes]
            Axes to plot on.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        CADETProcess.plotting
        """
        if self.npar is None:
            ax = self._plot_1D(ax, t, vmax)
        else:
            if comp is None and self.n_comp > 1:
                raise ValueError("Must specify component index.")

            ax = self._plot_2D(ax, t, comp, vmax)

        return ax


class SolutionSolid(SolutionBase):
    """
    Solution in the particle solid phase of the ``UnitOperation``.

    Dimensions: NCOL * NRAD * sum_{j}^{NPARTYPE}{NBOUND,j * NPAR,j}
    """

    dimensions = SolutionBase.dimensions + [
        "axial_coordinates",
        "radial_coordinates",
        "particle_coordinates",
        "component_coordinates",
    ]

    def __init__(
        self,
        name: str,
        component_system: ComponentSystem,
        bound_states: list[int],
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
        axial_coordinates: Optional[npt.ArrayLike] = None,
        radial_coordinates: Optional[npt.ArrayLike] = None,
        particle_coordinates: Optional[npt.ArrayLike] = None,
    ) -> None:
        """
        Initialize the SolutionSolid.

        Parameters
        ----------
        name : str
            The name of the solution object.
        component_system : ComponentSystem
            The component system associated with the solution.
        bound_states: list[int]
            The number of bound states per component.
        time : npt.ArrayLike
            An array-like structure representing time points.
        solution : npt.ArrayLike
            An array-like structure representing the solution data.
        axial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing axial coordinates.
            If None, it is assumed that the model has a singleton axial dimension.
        radial_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing radial coordinates.
            If None, it is assumed that the model has a singleton bulk dimension.
        particle_coordinates : Optional[npt.ArrayLike],
            An array-like structure representing particle coordinates.
            If None, it is assumed that the model has a singleton particle dimension.
        """
        self.bound_states = bound_states

        self.axial_coordinates = axial_coordinates
        self.radial_coordinates = radial_coordinates
        self.particle_coordinates = particle_coordinates

        super().__init__(name, component_system, time, solution)

    @property
    def n_bound(self) -> int:
        """int: Number of bound states."""
        return sum(self.bound_states)

    @property
    def ncol(self) -> int:
        """int: Number of axial discretization points."""
        if self.axial_coordinates is None:
            return None
        else:
            return len(self.axial_coordinates)

    @property
    def nrad(self) -> int:
        """int: Number of radial discretization points."""
        if self.radial_coordinates is None:
            return
        return len(self.radial_coordinates)

    @property
    def npar(self) -> int:
        """int: Number of particle discretization points."""
        if self.particle_coordinates is None:
            return
        return len(self.particle_coordinates)

    @plotting.create_and_save_figure
    def plot(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        components: Optional[list[str]] = None,
        layout: Optional[plotting.Layout] = None,
        y_max: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the entire solid phase solution for each component.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        components : Optional[list[str]]
            List of components to be plotted. If None, all components are plotted.
        layout : Optional[plotting.Layout]
            Plot layout options.
        y_max : Optional[float]
            Maximum value of y axis.
            If None, value is automatically deferred from solution.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes. The default is True.
        ax : Optional[Axes]
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        ax : Axes
            Axes object with concentration profile.

        Raises CADETProcessError
            If solution is not 1D.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plotlib
        plot_purity
        """
        if not (self.ncol is None and self.nrad is None and self.npar is None):
            raise CADETProcessError(
                "Solution has more than single dimension. Please use `plot_at_time`."
            )

        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [start, end]},
        )

        x = solution.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$c~/~mM$"
            layout.x_lim = (start, end)
        if y_max is not None:
            layout.y_lim = (None, y_max)

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        return ax

    def _plot_1D(
        self,
        t: float,
        components: list[str] | None = None,
        layout: plotting.Layout | None = None,
        ax: Axes | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot solid solution over space at given time.

        Parameters
        ----------
        t : float
            Time for plotting, in seconds.
            If t == -1, the final solution is plotted.
        components : list, optional
            List of components to be plotted. If None, all components are plotted.
            The default is None.
        layout : plotting.Layout, optional
            Plot layout options. If None, a new instance is created.
            The default is None.
        ax : Axes
            Axes to plot on.
        *args : Any
            Optional arguments passed down to _plot_solution_1D.
        **kwargs : Any
            Optional arguments passed down to _plot_solution_1D.

        Returns
        -------
        ax : Axes
            Axes object with concentration profile.

        See Also
        --------
        _plot_solution_1D
        slice_solution
        plot_at_position
        plotlib
        """
        solution = slice_solution(
            self,
            components=components,
            use_total_concentration=False,
            use_total_concentration_components=False,
            coordinates={"time": [t, t]},
        )

        x = self.axial_coordinates

        if layout is None:
            layout = plotting.Layout()
            layout.x_label = "$z~/~m$"
            layout.y_label = "$c~/~mM$"

        ax = _plot_solution_1D(ax, x, solution, layout, *args, **kwargs)

        plotting.add_text(ax, f"time = {t:.2f} s")

        return ax

    def _plot_2D(
        self,
        ax: Axes,
        t: float,
        comp: int,
        bound_state: int = 0,
        vmax: Optional[float] = None,
    ) -> tuple[Axes, QuadMesh]:
        x = self.axial_coordinates
        y = self.particle_coordinates

        if not self.time[0] <= t <= self.time[-1]:
            raise ValueError("Time exceeds bounds.")
        t_i = np.where(t <= self.time)[0][0]
        c_i = np.sum(self.bound_states[0:comp]) + bound_state
        v = self.solution[t_i, :, :, c_i].transpose()

        if vmax is None:
            vmax = v.max()

        mesh = ax.pcolormesh(x, y, v, shading="gouraud", vmin=0, vmax=vmax)

        plotting.add_text(ax, f"time = {t:.2f} s")

        layout = plotting.Layout()
        layout.title = (
            f"Solid phase concentration, comp={comp}, bound_state={bound_state}"
        )
        layout.x_label = "$z~/~m$"
        layout.y_label = "$r~/~m$"
        layout.labels = self.component_system.species[c_i]
        plotting.set_layout(ax, layout)

        return ax, mesh

    @plotting.create_and_save_figure
    def plot_at_time(
        self,
        t: float,
        comp: Optional[int] = None,
        bound_state: Optional[int] = 0,
        vmax: Optional[float] = None,
        ax: Optional[Axes] = None,
    ) -> Axes:
        """
        Plot particle solid solution for a given bound state over space at given time.

        Parameters
        ----------
        t : float
            Solution time at with to plot.
        comp : Optional[int], default=None
            Number of components.
        bound_state : int, default=0
            Bound state index.
        vmax: Optional[float]
            Maximum data value to scale color map.
        ax : Optional[Axes]
            Axes to plot on.

        Returns
        -------
        Axes
            Axes object with concentration profile.

        See Also
        --------
        CADETProcess.plotting
        """
        if self.npar is None:
            ax = self._plot_1D(ax, t, vmax)
        else:
            if comp is None and self.n_comp > 1:
                raise ValueError("Must specify component index.")
            if comp is None:
                comp = 0
            if bound_state is None and self.bound_states[comp] > 1:
                raise ValueError("Must specify bound state index.")

            ax, mesh = self._plot_2D(ax, t, comp, bound_state, vmax=vmax)
            plt.colorbar(mesh, ax)
        return ax


class SolutionVolume(SolutionBase):
    """Volume solution (of e.g. CSTR)."""

    @property
    def solution_shape(self) -> tuple[int, ...]:
        """tuple: (Expected) shape of the solution."""
        return (self.nt,)

    @plotting.create_and_save_figure
    def plot(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
        x_axis_in_minutes: Optional[bool] = True,
        ax: Optional[Axes] = None,
        update_layout: Optional[bool] = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the unit operation"s volume over time.

        Parameters
        ----------
        start : Optional[float]
            Start time for plotting in seconds. If None is provided, the first data
            point will be used as the start time. The default is None.
        end : Optional[float]
            End time for plotting in seconds. If None is provided, the last data point
            will be used as the end time. The default is None.
        x_axis_in_minutes : Optional[bool], default=True
            If True, the x-axis will be plotted using minutes.
        ax : Optional[Axes]
            Axes to plot on.
        update_layout : Optional[bool], default=True
            If True, update the figure"s layout.
        **kwargs : Any
            Additional arguments passed down to ax.plot()

        Returns
        -------
        Axes
            Axes object with the plot.

        See Also
        --------
        CADETProcess.plot
        """
        x = self.time
        if x_axis_in_minutes:
            x = x / 60
            if start is not None:
                start = start / 60
            if end is not None:
                end = end / 60

        y = self.solution * 1000

        y_min = np.min(y)
        y_max = 1.1 * np.max(y)

        layout = plotting.Layout()
        layout.x_label = "$time~/~s$"
        if x_axis_in_minutes:
            layout.x_label = "$time~/~min$"
        layout.y_label = "$V~/~L$"
        layout.x_lim = (start, end)
        layout.y_lim = (y_min, y_max)
        ax.plot(x, y, **kwargs)

        if update_layout:
            plotting.set_layout(ax, layout)

        return ax


def _plot_solution_1D(
    ax: Axes,
    x: np.ndarray,
    solution: SolutionBase,
    layout: Optional[Any] = None,
    plot_species: bool = False,
    plot_components: bool = True,
    plot_total_concentration: bool = False,
    alpha: int = 1,
    hide_labels: bool = False,
    hide_species_labels: bool = False,
    secondary_axis: Optional[Axes] = None,
    secondary_layout: Optional[Any] = None,
    show_legend: bool = True,
    update_layout: bool = True,
) -> Axes:
    sol = solution.solution
    c_total_comp = solution.total_concentration_components

    if secondary_axis is not None:
        ax_secondary = ax.twinx()
    else:
        ax_secondary = None

    species_index = 0
    y_min = 0
    y_max = 0
    y_min_sec = 0
    y_max_sec = 0
    colors = iter(plt.rcParams["axes.prop_cycle"].by_key()["color"])

    for i, comp in enumerate(solution.component_system.components):
        color = next(colors)
        if hide_labels or not update_layout:
            label = None
        else:
            label = comp.name

        if secondary_axis is not None and comp.name in secondary_axis.components:
            a = ax_secondary
        else:
            a = ax

        if plot_components:
            y_comp = c_total_comp[..., i]
            if secondary_axis is not None and comp.name in secondary_axis.components:
                if secondary_axis.transform is not None:
                    y_comp = secondary_axis.transform(y_comp)
                y_min_sec = min(min(y_comp), y_min_sec)
                y_max_sec = max(max(y_comp), y_max_sec)
            else:
                y_min = np.min((np.min(y_comp), y_min))
                y_max = np.max((np.max(y_comp), y_max))

            y_comp = np.squeeze(y_comp)

            a.plot(
                x,
                y_comp,
                linestyle="-",
                color=color,
                alpha=alpha,
                label=label,
            )

        if plot_species:
            if comp.n_species == 1:
                if plot_components:
                    species_index += 1
                    continue

            linestyle_iter = iter(plotting.linestyle_cycler)

            for s, species in enumerate(comp.label):
                if hide_species_labels or not update_layout:
                    label = None
                else:
                    label = species

                if (
                    comp.n_species == 1
                    and not plot_total_concentration
                    and not plot_components
                ):
                    linestyle = "-"
                else:
                    linestyle = next(linestyle_iter)["linestyle"]

                if (
                    secondary_axis is not None
                    and secondary_axis.transform is not None
                    and comp.name in secondary_axis.components
                ):
                    y_spec = secondary_axis.transform(sol[..., species_index])
                    y_min_sec = min(min(y_spec), y_min_sec)
                    y_max_sec = max(max(y_spec), y_max_sec)
                else:
                    y_spec = sol[..., species_index]
                    y_min = min(min(y_spec), y_min)
                    y_max = max(max(y_spec), y_max)

                y_spec = np.squeeze(y_spec)

                a.plot(
                    x,
                    y_spec,
                    linestyle=linestyle,
                    label=label,
                    color=color,
                    alpha=alpha,
                )
                species_index += 1

    if plot_total_concentration:
        if hide_labels or not update_layout:
            label = None
        else:
            label = "Total concentration"

        y_total = solution.total_concentration
        y_total = np.squeeze(y_total)
        y_min = min(min(y_total), y_min)
        y_max = max(max(y_total), y_max)
        a.plot(
            x,
            y_total,
            "-",
            label=label,
            color="k",
            alpha=alpha,
        )

    if layout.y_lim is None:
        layout.y_lim = (y_min, 1.1 * y_max)

    if secondary_axis is not None and secondary_layout is None:
        secondary_layout = plotting.Layout()
        secondary_layout.y_label = secondary_axis.y_label
        secondary_layout.y_lim = (y_min_sec, 1.1 * y_max_sec)

    if update_layout:
        plotting.set_layout(
            ax,
            layout,
            show_legend,
            ax_secondary,
            secondary_layout,
        )

    return ax


class InterpolatedSignal:
    def __init__(self, time: np.ndarray, signal: npt.ArrayLike) -> None:
        if len(signal.shape) == 1:
            signal = np.array(signal, ndmin=2).transpose()
        self._solutions = [
            PchipInterpolator(time, signal[:, comp]) for comp in range(signal.shape[1])
        ]
        self._derivatives = [signal.derivative() for signal in self._solutions]
        self._antiderivatives = [signal.antiderivative() for signal in self._solutions]

    @property
    def solutions(self) -> list[PchipInterpolator]:
        """list[PchipInterpolator]: Interpolators for each component."""
        return self._solutions

    @property
    def derivatives(self) -> list[PchipInterpolator]:
        """list[PchipInterpolator]: Derivative interpolators for each component."""
        return self._derivatives

    def derivative(self, time: np.ndarray) -> np.ndarray:
        """
        Calculate derivatives of the solution spline(s) at given time points.

        Parameters
        ----------
        time : np.ndarray
            The time points to evaluate the derivatives at.

        Returns
        -------
        np.ndarray
            Derivatives of the solution spline(s) at given time.
        """
        der = np.empty((len(time), len(self.solutions)))
        for comp, der_i in enumerate(self.derivatives):
            der[:, comp] = der_i(time)

        return der

    @property
    def antiderivatives(self) -> list[PchipInterpolator]:
        """list[PchipInterpolator]: Antiderivative interpolators for each component."""
        return self._antiderivatives

    def antiderivative(self, time: np.ndarray) -> np.ndarray:
        """
        Antiderivative of the solution of the spline(s) at given time.

        Parameters
        ----------
        time : np.ndarray
            The time points to evaluate the antiderivatives at.

        Returns
        -------
        np.ndarray
            Antiderivatives of the solution spline(s) at given time.
        """
        anti = np.empty((len(time), len(self.solutions)))
        for comp, anti_i in enumerate(self.antiderivatives):
            anti[:, comp] = anti_i(time)

        return anti

    def integral(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None,
    ) -> np.ndarray:
        """
        Definite integral between start and end.

        Parameters
        ----------
        start : Optional[float]
            Lower integration bound.
        end: Optional[float]
            Upper integration bound.

        Returns
        -------
        np.ndarray
            Definite integral of the solution spline(s) between limits.
        """
        return np.array([comp.integrate(start, end) for comp in self._solutions])

    def __call__(self, t: float) -> np.ndarray:
        return np.array([comp(t) for comp in self._solutions]).transpose()


def slice_solution(
    solution_original: SolutionBase,
    components: Optional[str | list[str]] = None,
    use_total_concentration: Optional[bool] = False,
    use_total_concentration_components: Optional[bool] = False,
    coordinates: Optional[dict[str, Optional[tuple[Optional[int], Optional[int]]]]] = None,
) -> SolutionBase:
    """
    Slice a `Solution` object along specified dimensions, components or both.

    Parameters
    ----------
    solution_original : SolutionBase
        The `Solution` object to slice.
    components : Optional[str | list[str]]
        The names of the components to keep in the sliced `Solution`.
        If `None`, all components are kept. The default is None.
    use_total_concentration : Optional[bool], default=False
        If `True`, only the total concentration data is kept in the sliced `Solution`.
    use_total_concentration_components : Optional[bool], default=False
        If `True`, the total concentration data is kept for each individual species
        of each component in the sliced `Solution`.
    coordinates : Optional[dict[str, Optional[tuple[Optional[int], Optional[int]]]]]
        A dictionary mapping dimensions to slice coordinates. Each dimension in the
        solution object is represented by a key in the dictionary, and the corresponding
        value is a tuple of two elements specifying the start and end coordinates of the
        slice along that dimension.
        If a value is `None`, the corresponding coordinate is not sliced.

    Returns
    -------
    SolutionBase
        A new solution object representing the sliced data.

    Raises
    ------
    ValueError
        If any of the slice coordinates exceeds the bounds of its corresponding
        dimension.
    CADETProcessError
        If any of the specified components or dimensions does not exist in the original
        solution.
    """
    solution = copy.deepcopy(solution_original)

    slices = ()
    if coordinates is not None:
        coordinates = copy.deepcopy(coordinates)
        for i, (dim, coord) in enumerate(solution.coordinates.items()):
            if dim == "component_coordinates":
                continue
            if dim in coordinates:
                sl = list(coordinates.pop(dim))

                # Update slice bounds if they are None
                if sl[0] is None:
                    sl[0] = coord[0]
                if sl[1] is None:
                    sl[1] = coord[-1]

                # Check bounds
                if not (coord[0] <= sl[0] <= sl[1] <= coord[-1]):
                    raise ValueError(f"{dim} coordinates exceed bounds.")

                # Calculate start and end indices using searchsorted
                start_index = np.searchsorted(coord, sl[0], side="right") - 1
                end_index = np.searchsorted(coord, sl[1], side="left") + 1

                # Ensure only a single entry is returned if start and end elements are the same
                if sl[0] == sl[1]:
                    end_index = start_index + 1

                # Create slice and update solution
                sl = slice(start_index, end_index)
                slices += (sl,)
                setattr(solution, dim, coord[sl])
            else:
                slices += (slice(None),)

        if len(coordinates) > 0:
            raise CADETProcessError(f"Unknown dimensions: {coordinates.keys()}")

        solution.solution = solution_original.solution[slices]

    # First calculate total_concentration of components, if required.
    if use_total_concentration_components:
        sol_total_concentration_comp = solution.total_concentration_components
        for comp in solution.component_system:
            if comp.n_species > 1:
                comp._species = []
                comp.add_species(comp.name)
        solution.solution = sol_total_concentration_comp

    # Then, slice components. Note that component index can only be used if total
    # concentration of components has already been calculated and set as solution array.
    if components is not None:
        if not isinstance(components, list):
            components = [components]
        components = copy.deepcopy(components)

        component_system = copy.deepcopy(solution.component_system)
        component_indices = []
        for i, (name, component) in enumerate(
            solution.component_system.components_dict.items()
        ):
            if name not in components:
                component_system.remove_component(component.name)
            else:
                name = str(name)
                components.remove(name)
                if use_total_concentration_components:
                    component_indices.append(i)
                else:
                    component_indices += solution.component_system.indices[component.name]

        if len(components) != 0:
            raise CADETProcessError(f"Unknown components: {components}")

        solution_components = solution.solution[..., component_indices]
        solution.component_system = component_system
        solution.solution = solution_components

    # Only calculate total concentration after removing unwanted components.
    if use_total_concentration:
        solution_total_concentration = solution.total_concentration
        solution.component_system = ComponentSystem(["total_concentration"])
        solution.solution = solution_total_concentration

    solution.update_solution()

    return solution


def slice_solution_front(
    solution_original: SolutionIO,
    min_percent: Optional[float] = 0.02,
    max_percent: Optional[float] = 0.98,
    use_max_slope: Optional[bool] = False,
    return_indices: Optional[bool] = False,
) -> SolutionIO | tuple[SolutionIO, int, int]:
    """
    Slice the front of a given solution.

    Parameters
    ----------
    solution_original : SolutionIO
        The `Solution` object to slice.
    min_percent : Optional[float]
        Minimum percentage of the peak height to consider. Default is 0.02.
    max_percent : Optional[float]
        Maximum percentage of the peak height to consider. Default is 0.98.
    use_max_slope : Optional[bool]
        If True, cut use the maximum slope as end. Default if False.
    return_indices : Optional[bool]
        If True, also return the start end end indices of the front. Default if False.

    Returns
    -------
    SolutionIO or tuple[SolutionIO, int, int]
        The sliced `SolutionIO` object. If `return_indices` is True, also returns the
        start and end indices of the slice.
    """
    solution = copy.deepcopy(solution_original)

    # Initialize new sequence with zeros
    solution.solution[:] = 0

    # Determine whether to use derivative array
    solution_array = solution_original.solution
    if use_max_slope:
        solution_array = solution_original.derivative.solution

    # Determine the max value and its index
    max_value = np.max(solution_array)
    max_index = np.argmax(solution_array)

    # Determine the min and max percent values
    select_min = min_percent * max_value
    select_max = max_percent * max_value

    # Find the indices for slicing using logical indexing
    idx_min = np.where(solution_array[:max_index] <= select_min)[0][-1]
    idx_max = np.where(solution_array[:max_index] >= select_max)[0][0]

    # Slice the sequence
    solution.solution[idx_min : idx_max + 1] = \
        solution_original.solution[idx_min : idx_max + 1]
    solution.update_solution()

    if not return_indices:
        return solution

    return solution, idx_min, idx_max



"""
=========================================================
CADETProcessError (:mod:`CADETProcess.CADETProcessError`)
=========================================================

.. currentmodule:: CADETProcess.CADETProcessError

Exceptions in **CADET-Process**.

.. autosummary::
    :toctree: generated/

    CADETProcessError

"""  # noqa


class CADETProcessError(Exception):
    """Exception for errors in CADET-Process."""



"""
=================================
Logging (:mod:`CADETProcess.log`)
=================================

.. currentmodule:: CADETProcess.log

The CADETProcess.log module provides functionality for logging events in CADET-Process.

.. autosummary::
    :toctree: generated/

    loggers
    get_logger

"""  # noqa

import logging
import time
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Optional

import pathos

__all__ = ["loggers", "get_logger"]


LOG_FORMAT = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s")

loggers = {}


def get_logger(name: str, level: Optional[str] = None) -> logging.Logger:
    """
    Retrieve logger from loggers dictionary. Create new one if it does not already exist.

    Parameters
    ----------
    name : str
        The name of the logger.
    level : str, optional
        The logging level to set on the logger.

    Returns
    -------
    logging.Logger
        The logger object.
    """
    try:
        logger = loggers[name]
    except KeyError:
        logger = pathos.logger()
        loggers[name] = logger

    if level is not None:
        level = getattr(logging, level)
        logger.setLevel(level)

    return logger


def update_loggers(log_directory: str, save_log: bool) -> None:
    """
    Update the file handlers of all logger objects in the loggers dictionary.

    Parameters
    ----------
    log_directory : str
        The directory to store the log files.
    save_log : bool
        If True, log files are saved. Otherwise, no files are saved.
    """
    for name, logger in loggers.items():
        update_file_handlers(log_directory, logger, name, save_log)


def update_file_handlers(
    log_directory: str,
    logger: logging.Logger,
    name: str,
    save_log: Optional[bool],
) -> None:
    """
    Update the file handlers of a logger object.

    Parameters
    ----------
    log_directory : str
        The directory to store the log files.
    logger : logging.Logger
        The logger object to update.
    name : str
        The name of the logger.
    save_log : bool
        If True, log files are saved. Otherwise, no files are saved.
    """
    try:
        level = logger.handlers[0].level
    except IndexError:
        level = logger.level

    for hdlr in logger.handlers:
        logger.removeHandler(hdlr)

    if save_log:
        add_file_handler(log_directory, logger, name, level)


def add_file_handler(
    log_directory: str,
    logger: logging.Logger,
    name: str,
    level: str,
    overwrite: Optional[bool] = False,
) -> None:
    """
    Add a file handler to a logger object.

    Parameters
    ----------
    log_directory : str
        The directory to store the log files.
    logger : logging.Logger
        The logger object to update.
    name : str
        The name of the logger.
    level : str
        The logging level to set on the logger.
    overwrite : bool, optional
        If True, the log file is overwritten. Otherwise, logs are appended to the file.
    """
    log_directory = Path(log_directory)
    log_directory.mkdir(exist_ok=True)

    if overwrite:
        mode = "w"
    else:
        mode = "a"

    file_handler = logging.FileHandler(log_directory / f"{name}.log", mode=mode)
    file_handler.setFormatter(LOG_FORMAT)
    file_handler.setLevel(level)

    logger.addHandler(file_handler)


def log_time(logger_name: str, level: Optional[int] = None) -> Callable:
    """
    Log execution time of function.

    Parameters
    ----------
    logger_name : str, optional
        Name of the logger.
    level : int, optional
        Log level.
    """

    def log_time_decorator(function: Callable) -> Any:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            start = time.time()
            result = function(*args, **kwargs)
            elapsed = time.time() - start
            logger = get_logger(logger_name, level=None)
            logger.debug(f"Execution of {str(function)} took {elapsed} s")
            return result

        return wrapper

    return log_time_decorator


def log_exceptions(logger_name: str, level: Optional[int] = None) -> Callable:
    """
    Log exceptions.

    Parameters
    ----------
    logger_name : str
        Name of the logger.
    level : int
        Log level.
    """

    def log_exception_decorator(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            logger = get_logger(logger_name, level=None)
            try:
                return function(*args, **kwargs)
            except Exception as e:
                # log the exception
                err = "There was an exception in "
                err += function.__name__
                logger.exception(err)

                # re-raise the exception
                raise e

        return wrapper

    return log_exception_decorator


def log_results(logger_name: str, level: Optional[int] = None) -> Callable:
    """
    Log results.

    Parameters
    ----------
    logger_name : str
        Name of the logger
    level : int
        Log level.
    """

    def log_results_decorator(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            logger = get_logger(logger_name, level=None)

            logger.debug("{} was called with {}, {}".format(function, *args, **kwargs))
            results = function(*args, **kwargs)
            logger.debug(f"Results: {results}")

            return results

        return wrapper

    return log_results_decorator



import unittest

# Import the functions to test
from CADETProcess.dataStructure import deprecated_alias, rename_kwargs


class TestDeprecatedAlias(unittest.TestCase):
    def test_basic_functionality(self):
        @deprecated_alias(old_arg="new_arg")
        def example_func(new_arg):
            return new_arg

        # Test with new argument name
        self.assertEqual(example_func(new_arg=42), 42)

        # Test with old argument name
        with self.assertWarns(DeprecationWarning) as warning_context:
            result = example_func(old_arg=42)
            self.assertEqual(result, 42)

        # Verify warning message
        warning_message = str(warning_context.warning)
        self.assertIn("`old_arg` is deprecated", warning_message)
        self.assertIn("use `new_arg` instead", warning_message)

    def test_argument_conflict(self):
        """Test error when both old and new argument names are used."""

        @deprecated_alias(old_arg="new_arg")
        def example_func(new_arg):
            return new_arg

        # Should raise TypeError when both old and new names are used
        with self.assertRaises(TypeError) as context:
            example_func(old_arg=1, new_arg=2)

        self.assertIn(
            "received both old_arg and new_arg as arguments", str(context.exception)
        )

    def test_rename_kwargs_direct(self):
        """Test rename_kwargs function directly."""
        kwargs = {"old_arg": 42}
        aliases = {"old_arg": "new_arg"}

        with self.assertWarns(DeprecationWarning):
            rename_kwargs("test_func", kwargs, aliases)

        self.assertIn("new_arg", kwargs)
        self.assertNotIn("old_arg", kwargs)
        self.assertEqual(kwargs["new_arg"], 42)

    def test_positional_args(self):
        """Test that positional arguments work normally with the decorator."""

        @deprecated_alias(old_kwarg="new_kwarg")
        def example_func(pos_arg, new_kwarg=None):
            return f"{pos_arg}-{new_kwarg}"

        # Test with positional argument
        self.assertEqual(example_func("test", new_kwarg=42), "test-42")

        # Test with old keyword argument
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(example_func("test", old_kwarg=42), "test-42")

    def test_no_arguments(self):
        """Test function with no arguments still works with decorator."""

        @deprecated_alias(old_arg="new_arg")
        def example_func():
            return "success"

        self.assertEqual(example_func(), "success")


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from typing import Any

import numpy as np


class MetricBase(ABC):
    """
    Abstract base class for metrics used in optimization, fractionation, and comparison.

    Attributes
    ----------
    n_metrics : int
        Number of metrics, default is 1.
    bad_metrics : float
        Value representing a bad metric, default is infinity.

    """

    n_metrics = 1
    bad_metrics = np.inf

    @abstractmethod
    def evaluate(self) -> Any:
        """
        Evaluate the metric.

        This method should be implemented by subclasses to perform specific metric calculations.
        """
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Invoke the evaluate method.

        Parameters
        ----------
        *args : tuple
            Variable length argument list passed to evaluate.
        **kwargs : dict
            Arbitrary keyword arguments passed to evaluate.

        Returns
        -------
        Any
            The result of the metric evaluation.
        """
        return self.evaluate(*args, **kwargs)

    def __str__(self) -> str:
        """
        Return the class name as the string representation of the instance.

        Returns
        -------
        str
            The name of the class.
        """
        return self.__class__.__name__



import multiprocessing
from abc import abstractmethod
from collections.abc import Iterable

from CADETProcess.dataStructure import (
    Constant,
    RangedInteger,
    Structure,
    UnsignedInteger,
)

cpu_count = multiprocessing.cpu_count()

__all__ = ["ParallelizationBackendBase", "SequentialBackend"]


class ParallelizationBackendBase(Structure):
    """
    Base class for all parallelization backend adapters.

    Attributes
    ----------
    n_cores : int
        Number of cores to be used. If set to 0 or -1, all available cores are used.
        For values less than -1, (n_cpus + 1 + n_cores) are used.
        For example, for n_cores = -2, all CPUs but one are used.
    """

    n_cores = RangedInteger(lb=-cpu_count, ub=cpu_count, default=1)

    _parameters = ["n_cores"]

    @property
    def _n_cores(self) -> int:
        if self.n_cores == 0:
            return cpu_count

        if self.n_cores < 0:
            return cpu_count + 1 + self.n_cores
        else:
            return self.n_cores

    @abstractmethod
    def evaluate(self, function: callable, population: Iterable) -> list:
        """
        Evaluate the function at all individuals in the population-list.

        This method must be implemented by subclasses to evaluate the provided function
        at each individual in the given population. The evaluation can be performed
        sequentially or in parallel, depending on the concrete implementation.

        Parameters
        ----------
        function : callable
            The function of interest to be evaluated for the population.
        population : Iterable
            A collection of individuals at which the function is to be evaluated.

        Returns
        -------
        list
            List of results of function evaluations.
        """
        pass

    def __str__(self) -> str:
        """Return the class name as a string."""
        return self.__class__.__name__


class SequentialBackend(ParallelizationBackendBase):
    """
    Sequential execution backend for evaluating the target function.

    This backend does not perform parallelization. It evaluates the function at each
    individual in the population sequentially, one by one.

    Attributes
    ----------
    n_cores : int
        Number of cores to be used. This attribute is constant and always set to 1
        for the 'SequentialBackend'.
    """

    n_cores = Constant(value=1)

    def evaluate(self, function: callable, population: Iterable) -> list:
        """
        Evaluate the function sequentially at all individuals in the population.

        Since this is the 'SequentialBackend', the evaluation is done sequentially
        without any parallelization.

        Parameters
        ----------
        function : callable
            The function of interest to be evaluated for the population.
        population : Iterable
            A collection of individuals at which the function is to be evaluated.

        Returns
        -------
        list
            List of results of function evaluations.
        """
        results = []
        for ind in population:
            results.append(function(ind))
        return results


try:
    from joblib import Parallel, delayed

    __all__.append("Joblib")
except ModuleNotFoundError:
    pass


class Joblib(ParallelizationBackendBase):
    """
    Parallelization backend implementation using joblib.

    Attributes
    ----------
    verbose : int
        The verbosity level: if nonzero, progress messages are printed.
        Above 50, the output is sent to stdout.
        The frequency of the messages increases with the verbosity level.
        If it's more than 10, all iterations are reported.
    """

    verbose = UnsignedInteger(default=0)
    _parameters = ["verbose"]

    def evaluate(self, function: callable, population: Iterable) -> list:
        """
        Evaluate the function in parallalel for all individuals of the population.

        Parameters
        ----------
        function : callable
            The function of interest to be evaluated for the population.
        population : Iterable
            A collection of individuals at which the function is to be evaluated.

        Returns
        -------
        list
            List of results of function evaluations.
        """
        backend = Parallel(n_jobs=self.n_cores, verbose=self.verbose)
        results = backend(delayed(function)(x) for x in population)

        return results


try:
    import pathos

    __all__.append("Pathos")
except ModuleNotFoundError:
    pass


class Pathos(ParallelizationBackendBase):
    """Parallelization backend using the pathos library."""

    def evaluate(self, function: callable, population: Iterable) -> list:
        """
        Evaluate the function in parallalel for all individuals of the population.

        Parameters
        ----------
        function : callable
            The function of interest to be evaluated for the population.
        population : Iterable
            A collection of individuals at which the function is to be evaluated.

        Returns
        -------
        list
            List of results of function evaluations.
        """
        with pathos.pools.ProcessPool(ncpus=self.n_cores) as pool:
            results = pool.map(function, population)
        return results



import copy
from collections import defaultdict
from typing import Optional

import numpy as np
import numpy.typing as npt
from matplotlib.axes import Axes

from CADETProcess import CADETProcessError, plotting
from CADETProcess.processModel import ComponentSystem, MassActionLaw, ReactionBaseClass


def preprocessing(
    reaction_system: ReactionBaseClass,
    buffer: npt.ArrayLike,
    pH: Optional[float | np.ndarray] = None,
    components: Optional[list] = None,
) -> tuple[dict[str, list], dict[str, np.ndarray], np.ndarray, dict[str, int], bool]:
    """
    Preprocess reaction system data for further analysis or simulation.

    This function prepares a reaction system by processing buffer concentrations,
    component indices, and pH values. It calculates pKa values for acids in the system
    and filters components based on user input.

    Parameters
    ----------
    reaction_system : ReactionBaseClass
        The reaction system containing components and reactions to be processed.
    buffer : npt.ArrayLike
        Buffer concentrations.
    pH : Optional[float | np.ndarray], optional
        pH value(s) for the system. If None, pH is calculated from the buffer concentrations.
    components : Optional[list], optional
        List of components to consider. If None, all components are considered.

    Returns
    -------
    tuple[dict[str, list], dict[str, np.ndarray], np.ndarray, dict[str, int], bool]
        A tuple containing:
        - pKa: Dictionary mapping component names to lists of pKa values.
        - c_acids_M: Dictionary mapping component names to their molar concentrations.
        - pH: Array of pH values.
        - indices: Dictionary mapping component names to their indices.
        - scalar_input: Boolean indicating if the input pH was scalar.

    Raises
    ------
    CADETProcessError
        If protons ('H+') are not found in the component system.
    """
    buffer = np.array(buffer, ndmin=2)
    buffer_M = 1e-3 * buffer

    component_system = copy.deepcopy(reaction_system.component_system)

    indices = component_system.indices
    if components is not None:
        for comp in indices.copy():
            if comp not in components:
                indices.pop(comp)

    try:
        proton_index = indices.pop("H+")
    except ValueError:
        raise CADETProcessError("Could not find proton in component system")

    if pH is None:
        pH = -np.log10(buffer_M[:, proton_index]).reshape((-1))
    else:
        pH = np.asarray(pH, dtype="float64")

    scalar_input = False
    if pH.ndim == 0:
        pH = pH[None]  # Makes x 1D
        scalar_input = True

    pKa = defaultdict(list)
    for r in reaction_system.reactions:
        reaction_indices = np.where(r.stoich)[0]
        for comp, i in indices.items():
            if not all(r_i in i + proton_index for r_i in reaction_indices):
                continue

            pKa[comp].append(-np.log10(r.k_eq * 1e-3))

    c_acids_M = {comp: buffer_M[:, i] for comp, i in indices.items() if comp in pKa}

    for comp in indices.copy():
        if comp not in pKa:
            indices.pop(comp)

    return pKa, c_acids_M, pH, indices, scalar_input


def c_species_nu(pKa: list[float], pH: float | list[float]) -> np.ndarray:
    """
    Compute normalized acid species concentration at given pH.

    Parameters
    ----------
    pKa : list
        List of pKa values.
    pH : float or list of floats.
        pH value

    Returns
    -------
    c_species_nu : np.ndarray
        Normalized acid species concentration.
    """
    pKa = np.array([1.0] + pKa)
    k_eq = 10 ** (-pKa)
    n = len(k_eq)

    c_H = np.power(10, -pH)
    c_species_nu = np.zeros((n, len(pH)))

    for j in range(n):
        k = np.prod(k_eq[0 : j + 1])
        c_species_nu[j] = k * c_H ** (n - j)

    return c_species_nu


def c_total_nu(pKa: list[float], pH: float | list[float]) -> np.ndarray:
    """
    Compute normalized total acid concentration at given pH.

    Parameters
    ----------
    pKa : list
        List of pKa values
    pH : float or list of floats.
        pH value

    Returns
    -------
    c_total_nu : np.ndarray
        Normalized acid species concentration.
    """
    return sum(c_species_nu(pKa, pH))


def z_total_nu(pKa: list[float], pH: float | list[float]) -> np.ndarray:
    """
    Compute normalized total charge at given pH.

    Parameters
    ----------
    pKa : list
        List of pKa values
    pH : float or list of floats.
        pH value

    Returns
    -------
    z_total_nu : np.ndarray
        Normalized acid species concentration.
    """
    c = c_species_nu(pKa, pH)

    return np.dot(np.arange(len(c)), c)


def eta(pKa: list[float], pH: float | list[float]) -> np.ndarray:
    """
    Compute degree of dissociation at given pH.

    Parameters
    ----------
    pKa : list
        List of pKa values
    pH : float or list of floats.
        pH value

    Returns
    -------
    eta : np.ndarray
        Degree of dissociation.
    """
    return z_total_nu(pKa, pH) / c_total_nu(pKa, pH)


def charge_distribution(
    reaction_system: ReactionBaseClass,
    pH: float | np.ndarray,
    components: Optional[list] = None,
) -> np.ndarray:
    """
    Calculate charge distribution at given pH.

    Parameters
    ----------
    reaction_system : ReactionModel
        Reaction system with deprotonation reactions.
    pH : float or array
        pH value of buffer.
    components : list, optional
        List of components to be considered in buffer capacity calculation.
        If None, all components are considered.

    Returns
    -------
    charge_distribution : np.ndarray
        Degree of protolysis; ratio of the concentration of the species to the
        total concentration.
    """
    buffer = reaction_system.n_comp * [1]
    pKa, c_acids_M, pH, indices, scalar_input = preprocessing(
        reaction_system, buffer, pH, components
    )

    if components is None:
        z_shape = (len(pH), reaction_system.n_comp - 1)
    else:
        n_comp = 0
        for comp in indices.values():
            n_comp += len(comp)
        z_shape = (len(pH), n_comp)

    z = np.zeros(z_shape)

    counter = 0
    for comp, ind in indices.items():
        z_comp = alpha(pKa[comp], pH)
        for j in range(len(ind)):
            z[:, counter] = z_comp[j, :]
            counter += 1

    if scalar_input:
        return np.squeeze(z)

    return z


def cummulative_charge_distribution(
    reaction_system: ReactionBaseClass,
    pH: float | np.ndarray,
    components: Optional[list] = None,
) -> np.ndarray:
    """
    Calculate cummulative charge at given pH.

    Parameters
    ----------
    reaction_system : ReactionModel
        Reaction system with deprotonation reactions.
    buffer : list
        Acid concentrations in mM.
    pH : float or array
        pH value of buffer.
    components : list, optional
        List of components to be considered in buffer capacity calculation.
        If None, all components are considerd.

    Returns
    -------
    cummulative_charge_distribution : np.ndarray
        Degree of dissociation;
    """
    buffer = reaction_system.n_comp * [1]
    pKa, c_acids_M, pH, indices, scalar_input = preprocessing(
        reaction_system, buffer, pH, components
    )

    z_cum = np.zeros((len(pH), len(indices)))

    for i, (comp, ind) in enumerate(indices.items()):
        charges = np.array(reaction_system.component_system.charges)[ind]
        max_charge = max(charges)
        z_cum[:, i] = max_charge - eta(pKa[comp], pH)

    if scalar_input:
        return np.squeeze(z_cum)

    return z_cum


def alpha(pKa: list, pH: float | list[float]) -> np.ndarray:
    """
    Compute degree of protolysis at given pH.

    Parameters
    ----------
    pKa : list
        List of pKa values
    pH : float or list of floats.
        pH value

    Returns
    -------
    alpha : np.ndarray
        Degree of protolysis.
    """
    return c_species_nu(pKa, pH) / c_total_nu(pKa, pH)


def beta(c_acid: type, pKa: list, pH: float | list[float]) -> np.ndarray:
    """
    Compute buffer capacity of acid at given pH.

    Parameters
    ----------
    c_acid : TYPE
        DESCRIPTION.
    pKa : list
        List of pKa values
    pH : float or list of floats.
        pH value

    Returns
    -------
    beta : np.ndarray
        Buffer capacity.
    """
    a = alpha(pKa, pH)
    beta = np.zeros(
        len(pH),
    )

    n = c_acid.shape[1]
    for j in range(1, n):
        for i in range(0, j):
            beta += (j - i) ** 2 * a[j] * a[i]

    beta *= np.log(10) * np.sum(c_acid, axis=1)

    return beta


def beta_water(pH: float | list[float]) -> np.ndarray:
    """
    Compute buffer capacity of water.

    Parameters
    ----------
    pH : float or list of floats.
        pH value

    Returns
    -------
    beta_water
        Buffer capacity of water.
    """
    c_H = 10 ** (-pH)
    return np.log(10) * (10 ** (-14) / c_H + c_H)


def buffer_capacity(
    reaction_system: ReactionBaseClass,
    buffer: list,
    pH: Optional[float | np.ndarray] = None,
    components: list = None,
) -> np.ndarray:
    """
    Calculate buffer capacity at given buffer concentration and pH.

    Parameters
    ----------
    reaction_system : ReactionModel
        Reaction system with deprotonation reactions.
    buffer : list
        Acid concentrations in mM.
    pH : float or array, optional
        pH value of buffer. If None, value is inferred from buffer entry.
    components : list, optional
        List of components to be considered in buffer capacity calculation.
        If None, all components are considerd.

    Returns
    -------
    buffer_capacity : np.ndarray
        Buffer capacity in mM for individual acid components.
        To get overall buffer capacity, component capacities must be summed up.
    """
    pKa, c_acids_M, pH, indices, scalar_input = preprocessing(
        reaction_system, buffer, pH, components
    )

    buffer_capacity = np.zeros((len(pH), len(c_acids_M) + 1))

    for i, comp in enumerate(indices):
        buffer_capacity[:, i] = beta(c_acids_M[comp], pKa[comp], pH)

    buffer_capacity[:, -1] = beta_water(pH)

    buffer_capacity *= 1e3
    if scalar_input:
        return np.squeeze(buffer_capacity)

    return buffer_capacity


def ionic_strength(component_system: ComponentSystem, buffer: list) -> np.ndarray:
    """
    Compute ionic strength.

    Parameters
    ----------
    buffer : list
        Buffer concentrations in mM.
    component_system : ComponentSystem
        Component system; must contain charges.

    Returns
    -------
    i : np.ndarray
        Ionic strength of buffer
    """
    if not isinstance(component_system, ComponentSystem):
        raise TypeError("Expected ComponentSystem")
    if len(buffer) != component_system.n_comp:
        raise CADETProcessError("Number of components does not match")

    buffer = np.asarray(buffer, dtype="float64")
    z = np.asarray(component_system.charges)
    return 1 / 2 * np.sum(buffer * z**2)


@plotting.create_and_save_figure
def plot_buffer_capacity(
    reaction_system: MassActionLaw,
    buffer: list,
    pH: Optional[np.ndarray] = None,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot buffer capacity of reaction system over pH at given concentration.

    Parameters
    ----------
    reaction_system : MassActionLaw
        Reaction system with stoichiometric coefficients and reaction rates.
    buffer : list
        Buffer concentration in mM.
    pH : Optional[np.ndarray]
        Range of pH to be plotted.
    ax : Optional[Axes]
        Axes to plot on.

    Returns
    -------
    ax : Axes
        Axes object with buffer capacity plot.
    """
    if pH is None:
        pH = np.linspace(0, 14, 101)

    b = buffer_capacity(reaction_system, buffer, pH)
    b_total = np.sum(b, axis=1)

    labels = reaction_system.component_system.names
    labels.remove("H+")

    for i in range(reaction_system.component_system.n_components - 1):
        ax.plot(pH, b[:, i], label=labels[i])

    ax.plot(pH, b[:, -1], label="Water")
    ax.plot(pH, b_total, "k--", label="Total buffer capacity")

    layout = plotting.Layout()
    layout.x_label = "$pH$"
    layout.y_label = "buffer capacity / mM"
    layout.y_lim = (0, 1.1 * np.max(b_total))

    plotting.set_layout(ax, layout)

    return ax


@plotting.create_and_save_figure
def plot_charge_distribution(
    reaction_system: MassActionLaw,
    pH: Optional[np.ndarray] = None,
    plot_cumulative: Optional[bool] = False,
    ax: Optional[Axes] = None,
) -> Axes:
    """
    Plot charge distribution of components over pH.

    Parameters
    ----------
    reaction_system : MassActionLaw
        Reaction system with stoichiometric coefficients and reaction rates.
    pH : Optional[np.ndarray]
        Range of pH to be plotted.
    plot_cumulative : Optional[bool]
        If True, only plot cumulative charge of each acid.
    ax : Optional[Axes]
        Axes to plot on.

    Returns
    -------
    ax : Axes
        Axes object with charge distribution plot.
    """
    if pH is None:
        pH = np.linspace(0, 14, 101)

    layout = plotting.Layout()

    if plot_cumulative:
        c = cummulative_charge_distribution(reaction_system, pH)
        layout.y_label = "degree of dissociation"
    else:
        c = charge_distribution(reaction_system, pH)
        layout.y_label = "degree of protolysis"

    if plot_cumulative:
        labels = reaction_system.component_system.names
    else:
        labels = reaction_system.component_system.species

    labels.remove("H+")

    for i, label in zip(c.T, labels):
        ax.plot(pH, i, label=label)

    layout.x_label = "$pH$"
    layout.y_lim = (1.1 * np.min(c), 1.1 * np.max(c))

    plotting.set_layout(ax, layout)

    return ax



from collections.abc import Mapping
from functools import reduce
from operator import getitem
from typing import Any, Generator, Sequence

__all__ = [
    "check_nested",
    "generate_nested_dict",
    "update_dict_recursively",
    "get_nested_value",
    "set_nested_value",
    "get_nested_attribute",
    "set_nested_attribute",
    "get_nested_list_value",
    "set_nested_list_value",
]


def check_nested(nested_dict: dict[str, Any], path: str | list) -> bool:
    """
    Check if a key path exists in a nested dictionary.

    Parameters
    ----------
    nested_dict : dict
        dictionary in which the path is checked.
    path : str or list
        Path to the key in dot notation (string) or as a list of keys.

    Returns
    -------
    bool
        True if the item exists and is not a dictionary, False otherwise.
    """
    if isinstance(path, str):
        path = path.split(".")

    try:
        value = get_nested_value(nested_dict, path)
        return not isinstance(value, dict)
    except (KeyError, TypeError):
        return False


def generate_nested_dict(path: str | list, value: Any = None) -> dict[str, Any]:
    """
    Generate a nested dictionary from a dot-separated path.

    Parameters
    ----------
    path : str or list
        Path to create in the dictionary, given in dot notation or as a list.
    value : Any, optional
        The value to assign to the innermost key.

    Returns
    -------
    dict
        A nested dictionary created from the path.
    """
    if isinstance(path, str):
        path = path.split(".")

    nested_dict = {path[-1]: value}
    for key in reversed(path[:-1]):
        nested_dict = {key: nested_dict}
    return nested_dict


def insert_path(nested_dict: dict[str, Any], path: str | list, value: Any) -> None:
    """
    Add a key path to an existing dictionary without overwriting existing keys.

    Parameters
    ----------
    nested_dict : dict
        dictionary to update.
    path : str or list
        Path to the key in dot notation (string) or as a list of keys.
    value : Any
        Value to insert.

    Raises
    ------
    KeyError
        If an intermediate key in the path does not exist.
    """
    if isinstance(path, str):
        path = path.split(".")

    if len(path) == 1:
        nested_dict.setdefault(path[0], value)
    else:
        nested_dict.setdefault(path[0], {})
        insert_path(nested_dict[path[0]], path[1:], value)


def get_leaves(nested_dict: dict[str, Any]) -> Generator[str, None, None]:
    """
    Yield leaf keys of a nested dictionary in dot notation.

    Parameters
    ----------
    nested_dict : dict
        The nested dictionary to traverse.

    Yields
    ------
    str
        The full path to each leaf node in dot notation.
    """
    for key, value in nested_dict.items():
        if isinstance(value, dict) and value:
            for subpath in get_leaves(value):
                yield f"{key}.{subpath}"
        else:
            yield key


def set_nested_value(nested_dict: dict[str, Any], path: str | list, value: Any) -> None:
    """
    Set a value in a nested dictionary using a dot-separated path.

    Parameters
    ----------
    nested_dict : dict
        dictionary to update.
    path : str or list
        Path to the key in dot notation (string) or as a list of keys.
    value : Any
        Value to set.

    Raises
    ------
    KeyError
        If an intermediate key in the path does not exist.
    """
    if isinstance(path, str):
        path = path.split(".")

    get_nested_value(nested_dict, path[:-1])[path[-1]] = value


def get_nested_value(nested_dict: dict[str, Any], path: str | list) -> Any:
    """
    Retrieve a value from a nested dictionary using a dot-separated path.

    Parameters
    ----------
    nested_dict : dict
        dictionary to retrieve the value from.
    path : str or list
        Path to the key in dot notation (string) or as a list of keys.

    Returns
    -------
    Any
        The value stored at the specified key path.

    Raises
    ------
    KeyError
        If the key path does not exist.
    """
    if isinstance(path, str):
        path = path.split(".")

    return reduce(getitem, path, nested_dict)


def update_dict_recursively(
    target_dict: dict[str, Any],
    updates: dict[str, Any],
    only_existing_keys: bool = False,
) -> dict[str, Any]:
    """
    Recursively update `target_dict` with values from `updates`.

    Parameters
    ----------
    target_dict : dict
        The original dictionary to be updated.
    updates : dict
        The dictionary containing new values to merge into `target_dict`.
    only_existing_keys : bool, optional
        If True, only update keys that already exist in `target_dict`.
        If False, add new keys from `updates`. Default is False.

    Returns
    -------
    dict
        The updated dictionary with `updates` applied.
    """
    for key, value in updates.items():
        if only_existing_keys and key not in target_dict:
            continue  # Skip keys that don't exist in target_dict

        if isinstance(value, Mapping) and isinstance(target_dict.get(key), Mapping):
            # Recursively update nested dictionaries
            target_dict[key] = update_dict_recursively(target_dict[key], value, only_existing_keys)
        else:
            # Directly update the value
            target_dict[key] = value

    return target_dict


def get_nested_attribute(obj: Any, path: str) -> Any:
    """
    Access a nested attribute using a dot-separated path.

    Parameters
    ----------
    obj : Any
        The object to retrieve the attribute from.
    path : str
        The dot-separated path to the nested attribute.

    Returns
    -------
    Any
        The value of the nested attribute.

    Raises
    ------
    AttributeError
        If any attribute in the path does not exist.
    """
    attributes = path.split(".")
    for attr in attributes:
        obj = getattr(obj, attr)
    return obj


def set_nested_attribute(obj: Any, attr_string: str, value: Any) -> None:
    """
    Set a nested attribute using a dot-separated path.

    Parameters
    ----------
    obj : Any
        The object to modify.
    attr_string : str
        The dot-separated path to the nested attribute.
    value : Any
        The value to set.
    """
    attributes = attr_string.split(".")
    for attr in attributes[:-1]:
        obj = getattr(obj, attr)
    setattr(obj, attributes[-1], value)


def get_nested_list_value(ls: Sequence[Any], idx_tuple: tuple[int, ...]) -> Any:
    """
    Retrieve a value from a nested list structure using an index tuple.

    Parameters
    ----------
    ls : Sequence[Any]
        The nested list structure.
    idx_tuple : tuple[int, ...]
        A tuple of indices specifying the path to the desired value.

    Returns
    -------
    Any
        The value at the specified nested position.

    Raises
    ------
    IndexError
        If any index in the path is out of range.
    """
    return reduce(lambda l, i: l[i], idx_tuple, ls)  # noqa: E741


def set_nested_list_value(
    ls: Sequence[Any],
    idx_tuple: tuple[int, ...],
    value: Any,
) -> None:
    """
    Set a value in a nested list structure using an index tuple.

    Parameters
    ----------
    ls : Sequence[Any]
        The nested list structure.
    idx_tuple : tuple[int, ...]
        A tuple of indices specifying the path to the value to be set.
    value : Any
        The new value to assign.

    Raises
    ------
    IndexError
        If any index in the path is out of range.
    """
    for idx in idx_tuple[:-1]:
        ls = ls[idx]  # Navigate through the nested lists
    ls[idx_tuple[-1]] = value  # Set the final value



"""
==================================================
Dynamic Events (:mod:`CADETProcess.dynamicEvents`)
==================================================

.. currentmodule:: CADETProcess.dynamicEvents

This module provides functionality for dynamic changes of parameters in CADET-Process.

Sections and Timelines
======================

.. autosummary::
    :toctree: generated/

    Section
    TimeLine
    MultiTimeLine

Events and EventHandlers
========================

.. autosummary::
    :toctree: generated/

    Event
    Duration
    EventHandler

"""

from .section import *
from .event import *



import unittest

import numpy
from CADETProcess.processModel import (
    BiLangmuir,
    BindingBaseClass,
    ComponentSystem,
    Langmuir,
    MobilePhaseModulator,
    MultistateStericMassAction,
    StericMassAction,
    binding,
)
from CADETProcess.simulator.cadetAdapter import adsorption_parameters_map


class Test_Binding(unittest.TestCase):
    def setUp(self):
        component_system = ComponentSystem(2)

        binding_model = Langmuir(component_system, name="test")

        binding_model.adsorption_rate = [0.02, 0.03]
        binding_model.desorption_rate = [1, 1]
        binding_model.capacity = [100, 100]

        self.langmuir = binding_model

        binding_model = BiLangmuir(component_system, name="test")

        binding_model.adsorption_rate = [0.02, 0.03, 0.001, 0.002]
        binding_model.desorption_rate = [1, 1, 2, 2]
        binding_model.capacity = [100, 100, 200, 200]

        self.bi_langmuir = binding_model

        binding_model = MultistateStericMassAction(component_system, name="test")

        self.multi_state_sma = binding_model

        binding_model = StericMassAction(component_system, name="test")

        binding_model.adsorption_rate = [0.02, 0.03]
        binding_model.desorption_rate = [1, 1]
        binding_model.characteristic_charge = [5, 10]
        binding_model.steric_factor = [10, 10]
        binding_model.capacity = 100
        binding_model.reference_liquid_phase_conc = 200
        binding_model.reference_solid_phase_conc = 100
        self.sma = binding_model

        binding_model = MobilePhaseModulator(component_system, name="test")

        binding_model.adsorption_rate = [0.02, 0.03]
        binding_model.desorption_rate = [1, 1]
        binding_model.capacity = [100, 100]
        binding_model.ion_exchange_characteristic = [1.4, 1.7]
        binding_model.hydrophobicity = [1.4, 1.7]
        binding_model.linear_threshold = 1e-8
        self.mobile_phase_modulator = binding_model

    def test_sma_reference_concentration_transformations(self):
        adsorption_rate = numpy.array(self.sma.adsorption_rate)
        desorption_rate = numpy.array(self.sma.desorption_rate)
        characteristic_charge = numpy.array(self.sma.characteristic_charge)
        expected_untransformed_adsorption_rate = adsorption_rate * (
            self.sma.reference_solid_phase_conc**-characteristic_charge
        )
        expected_untransformed_desorption_rate = desorption_rate * (
            self.sma.reference_liquid_phase_conc**-characteristic_charge
        )
        assert numpy.allclose(
            expected_untransformed_adsorption_rate,
            self.sma.adsorption_rate_untransformed,
        )
        assert numpy.allclose(
            expected_untransformed_desorption_rate,
            self.sma.desorption_rate_untransformed,
        )

        with self.assertRaises(ValueError):
            component_system = ComponentSystem(2)
            binding_model = StericMassAction(component_system, name="test")
            binding_model.adsorption_rate_untransformed = [0.02, 0.03]

    def test_in_cadetAdapter(self):
        binding_classes = {
            name: cls
            for name, cls in binding.__dict__.items()
            if (isinstance(cls, type) and issubclass(cls, BindingBaseClass))
        }
        binding_classes.pop("BindingBaseClass")

        for name, cls in binding_classes.items():
            self.assertIn(name, adsorption_parameters_map)
            map = adsorption_parameters_map[name]
            try:
                assert set(cls._parameters) == set(
                    map["parameters"].values()
                )  # This needs to compare to .values
            except AssertionError:
                raise AssertionError(
                    f"Isotherm '{name}' has these parameters in cadetAdapter.py missing: "
                    f"{set(cls._parameters) - set(map['parameters'].values())} "
                    "and these parameters in binding.py missing: "
                    f"{set(map['parameters'].values()) - set(cls._parameters)}"
                )

    def test_get_parameters(self):
        parameters_expected = {
            "is_kinetic": True,
            "adsorption_rate": [0.02, 0.03],
            "desorption_rate": [1.0, 1.0],
            "capacity": [100.0, 100.0],
        }
        parameters = self.langmuir.parameters
        self.assertDictEqual(parameters_expected, parameters)

    def test_binding_sites(self):
        self.assertEqual(self.bi_langmuir.n_binding_sites, 2)
        self.assertEqual(self.bi_langmuir.n_bound_states, 4)

        self.bi_langmuir.n_binding_sites = 3
        self.assertEqual(self.bi_langmuir.n_binding_sites, 3)
        self.assertEqual(self.bi_langmuir.n_bound_states, 6)

        with self.assertRaises(ValueError):
            self.langmuir.n_binding_sites = 2

    def test_bound_states(self):
        self.assertEqual(self.multi_state_sma.n_bound_states, 2)

        self.multi_state_sma.bound_states = [3, 2]
        self.assertEqual(self.multi_state_sma.n_bound_states, 5)

        with self.assertRaises(ValueError):
            self.multi_state_sma.bound_states = [1, 2, 3]

        with self.assertRaises(ValueError):
            self.langmuir.bound_states = [2, 2]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import copy
from abc import abstractmethod
from functools import wraps
from typing import Any, Callable, Optional, Union

import numpy as np
from scipy.integrate import simpson
from scipy.special import expit

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import UnsignedInteger
from CADETProcess.metric import MetricBase
from CADETProcess.reference import FractionationReference, ReferenceBase, ReferenceIO
from CADETProcess.solution import (
    SolutionBase,
    SolutionIO,
    slice_solution,
    slice_solution_front,
)

from .peaks import find_breakthroughs, find_peaks
from .shape import pearson, pearson_offset

__all__ = [
    "DifferenceBase",
    "calculate_sse",
    "calculate_rmse",
    "SSE",
    "RMSE",
    "NRMSE",
    "Norm",
    "L1",
    "L2",
    "AbsoluteArea",
    "RelativeArea",
    "Shape",
    "ShapeFront",
    "PeakHeight",
    "PeakPosition",
    "BreakthroughHeight",
    "BreakthroughPosition",
    "FractionationSSE",
]


def sigmoid_distance(
    measurement: Union[float, np.ndarray],
    target: Union[float, np.ndarray],
    normalization: float = 1,
) -> float | np.ndarray:
    """
    Calculate the distance between two values using a sigmoid function.

    The distance is defined as the sigmoid transformation of the difference between the
    measurement and target values, normalized by a user-specified factor.

    Parameters
    ----------
    measurement : float or numpy.ndarray
        The measurement value(s).
    target : float or numpy.ndarray
        The target value(s).
    normalization : float, optional
        The factor to use for normalization. Default is 1.

    Returns
    -------
    float or numpy.ndarray
        The sigmoid distance between the measurement and target values.

    Examples
    --------
    >>> sigmoid_distance(3, 5)
    0.7310585786300049

    >>> sigmoid_distance([1, 2, 3], [3, 2, 1])
    array([0.73105858, 0.5       , 0.26894142])
    """
    input = (measurement - target) / normalization
    output = np.abs(2 * expit(input) - 1)
    return output


class DifferenceBase(MetricBase):
    """
    Base class for difference metric evaluation between a reference and a solution.

    Parameters
    ----------
    reference : ReferenceBase
        Reference used for calculating difference metric.
    components : {str, list}, optional
        Solution components to be considered.
        If None, all components are considered. The default is None.
    use_total_concentration : bool, optional
        If True, use sum of all components. The default is False.
    use_total_concentration_components : bool, optional
        If True, sum concentration of species. The default is True.
    start : float, optional
        End time of solution slice to be considered. The default is None.
    end : float, optional
        End time of solution slice to be considered. The default is None.
    transform : callable, optional
        Function to transform solution. The default is None.
    resample : bool, optional
        If True, resample data. The default is True.
    smooth : bool, optional
        If True, smooth data. The default is False.
    normalize : bool, optional
        If True, normalize data. The default is False.
    """

    _valid_references: tuple[type] = ()

    def __init__(
        self,
        reference: ReferenceBase,
        components: Optional[dict[str, list]] = None,
        use_total_concentration: bool = False,
        use_total_concentration_components: bool = True,
        start: Optional[float] = None,
        end: Optional[float] = None,
        transform: Optional[Callable] = None,
        only_transforms_array: bool = True,
        resample: bool = True,
        smooth: bool = False,
        normalize: bool = False,
    ) -> None:
        """
        Initialize an instance of DifferenceBase.

        Parameters
        ----------
        reference : ReferenceBase
            Reference used for calculating difference metric.
        components : {str, list}, optional
            Solution components to be considered.
            If None, all components are considered. The default is None.
        use_total_concentration : bool, optional
            If True, use sum of all components. The default is False.
        use_total_concentration_components : bool, optional
            If True, sum concentration of species. The default is True.
        start : float, optional
            End time of solution slice to be considerd. The default is None.
        end : float, optional
            End time of solution slice to be considerd. The default is None.
        transform : callable, optional
            Function to transform solution. The default is None.
        only_transforms_array: bool, optional
            If True, only transform np array of solution object. The default is True.
        resample : bool, optional
            If True, resample data. The default is True.
        smooth : bool, optional
            If True, smooth data. The default is False.
        normalize : bool, optional
            If True, normalize data. The default is False.
        """
        self.components = components
        self.use_total_concentration = use_total_concentration
        self.use_total_concentration_components = use_total_concentration_components
        self.start = start
        self.end = end
        self.transform = transform
        self.only_transforms_array = only_transforms_array
        self.resample = resample
        self.smooth = smooth
        self.normalize = normalize
        self.reference = reference

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        return self.reference.solution.shape[-1]

    @property
    def bad_metrics(self) -> list:
        """list: Worst case values for each metric."""
        return self.n_metrics * [np.inf]

    @property
    def reference(self) -> SolutionBase:
        """SolutionBase: The reference solution."""
        return self._reference

    @reference.setter
    def reference(self, reference: SolutionBase) -> None:
        if not isinstance(reference, self._valid_references):
            raise TypeError(
                f"Invalid reference type: {type(reference)}. "
                f"Expected types: {self._valid_references}."
            )

        reference = copy.deepcopy(reference)
        self.reference_original = reference

        if self.resample:
            reference = reference.resample()
        if self.normalize:
            reference = reference.normalize()
        if self.smooth:
            reference = reference.smooth_data()

        reference = slice_solution(
            reference,
            self.components,
            self.use_total_concentration,
            self.use_total_concentration_components,
            coordinates={"time": (self.start, self.end)},
        )

        self._reference = reference

    def checks_dimensions(func: Callable) -> Callable:
        """Wrap method automatically check reference and solution dimensions."""

        @wraps(func)
        def wrapper(
            self: DifferenceBase,
            solution: SolutionBase,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Automatically check reference and solution dimensions before further processing."""
            if solution.solution.shape != self.reference.solution.shape:
                raise CADETProcessError("Simulation shape does not match experiment")
            value = func(self, solution, *args, **kwargs)
            return value

        return wrapper

    def slices_solution(func: Callable) -> Callable:
        """Wrap method to automatically slice solution."""

        @wraps(func)
        def wrapper(
            self: "DifferenceBase",
            solution: SolutionBase,
            slice: bool = True,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Automatically slice solution before further processing."""
            if slice:
                solution = slice_solution(
                    solution,
                    self.components,
                    self.use_total_concentration,
                    self.use_total_concentration_components,
                    coordinates={"time": (self.start, self.end)},
                )

            value = func(self, solution, *args, **kwargs)

            return value

        return wrapper

    def resamples_normalizes_and_smoothes_solution(func: Callable) -> Callable:
        """Wrap method to automatically resample, normalize, and smooth solution."""

        @wraps(func)
        def wrapper(
            self: DifferenceBase,
            solution: SolutionBase,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Automatically resample, normalize, and smooth solution before further processing."""
            solution = copy.deepcopy(solution)
            if self.resample:
                solution = solution.resample(
                    self.reference.time[0],
                    self.reference.time[-1],
                    len(self.reference.time),
                )
            if self.normalize:
                solution = solution.normalize()
            if self.smooth:
                solution = solution.smooth_data()

            value = func(self, solution, *args, **kwargs)
            return value

        return wrapper

    def transforms_solution(func: Callable) -> Callable:
        """Wrap method s.t. solution is automatically transformed."""

        @wraps(func)
        def wrapper(
            self: DifferenceBase,
            solution: SolutionBase,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """Automatically transforme solution data before further processing."""
            if self.transform is not None:
                solution = copy.deepcopy(solution)

                if self.only_transforms_array:
                    solution.solution = self.transform(solution.solution)
                else:
                    solution = self.transform(solution)

            value = func(self, solution, *args, **kwargs)
            return value

        return wrapper

    @abstractmethod
    def _evaluate(self) -> np.ndarray:
        """
        Abstract method to compute the difference metric.

        Returns
        -------
        np.ndarray
            The computed difference metric.
        """
        return

    @resamples_normalizes_and_smoothes_solution
    @slices_solution
    @transforms_solution
    @checks_dimensions
    def evaluate(self, solution: SolutionBase) -> np.ndarray:
        """
        Compute the difference between the reference solution and the input solution.

        Parameters
        ----------
        solution : Solution
            The solution to compare with the reference solution.

        Returns
        -------
        np.ndarray
            The difference between the two solutions.
        """
        metric = self._evaluate(solution)
        return metric

    __call__ = evaluate

    @resamples_normalizes_and_smoothes_solution
    @slices_solution
    @transforms_solution
    def slice_and_transform(self, solution: SolutionBase) -> SolutionBase:
        """
        Slice the solution and applies the transform callable (if defined).

        Parameters
        ----------
        solution : Solution
            The solution to slice and transform.

        Returns
        -------
        Solution
            The sliced and transformed solution.
        """
        return solution


def calculate_sse(simulation: np.ndarray, reference: np.ndarray) -> np.ndarray:
    """
    Calculate the sum of squared errors (SSE) between simulation and reference.

    Parameters
    ----------
    simulation : np.ndarray
        Array of simulated values.
    reference : np.ndarray
        Array of reference values.

    Returns
    -------
    np.ndarray
        The SSE between simulation and reference.
    """
    return np.sum((simulation - reference) ** 2, axis=0)


class SSE(DifferenceBase):
    """Sum of squared errors (SSE) difference metric."""

    _valid_references = (ReferenceIO, SolutionIO)

    def _evaluate(self, solution: SolutionBase) -> np.ndarray:
        sse = calculate_sse(solution.solution, self.reference.solution)

        return sse


def calculate_rmse(simulation: np.ndarray, reference: np.ndarray) -> np.ndarray:
    """
    Calculate the root mean squared error (RMSE) between simulation and reference.

    Parameters
    ----------
    simulation : np.ndarray
        Array of simulated values.
    reference : np.ndarray
        Array of reference values.

    Returns
    -------
    np.ndarray
        The RMSE between simulation and reference.
    """
    return np.sqrt(np.mean((simulation - reference) ** 2, axis=0))


class RMSE(DifferenceBase):
    """Root mean squared errors (RMSE) difference metric."""

    _valid_references = (SolutionIO, ReferenceIO)

    def _evaluate(self, solution: SolutionBase) -> np.ndarray:
        rmse = calculate_rmse(solution.solution, self.reference.solution)

        return rmse


class NRMSE(DifferenceBase):
    """Normalized root mean squared errors (RRMSE) difference metric."""

    _valid_references = (SolutionIO, ReferenceIO)

    def _evaluate(self, solution: SolutionBase) -> np.ndarray:
        rmse = calculate_rmse(solution.solution, self.reference.solution)
        nrmse = rmse / np.max(self.reference.solution, axis=0)

        return nrmse


class Norm(DifferenceBase):
    """
    Norm difference metric.

    Attributes
    ----------
    order : int
        The order of the norm.
    """

    _valid_references = (SolutionIO, ReferenceIO)

    order = UnsignedInteger()

    def _evaluate(self, solution: SolutionBase) -> float:
        norm = np.linalg.norm(
            (solution.solution - self.reference.solution), ord=self.order
        )

        return norm


class L1(Norm):
    """L1 norm difference metric."""

    order = 1


class L2(Norm):
    """L2 norm difference metric."""

    order = 2


class AbsoluteArea(DifferenceBase):
    """Absolute difference in area difference metric."""

    _valid_references = (SolutionIO, ReferenceIO)

    def _evaluate(self, solution: SolutionBase) -> Union[float, np.ndarray]:
        """
        np.ndarray: Absolute difference in area compared to reference.

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.
        """
        area_ref = simpson(self.reference.solution, self.reference.time, axis=0)
        area_sol = simpson(solution.solution, solution.time, axis=0)

        return abs(area_ref - area_sol)


class RelativeArea(DifferenceBase):
    """Relative difference in area difference metric."""

    _valid_references = (SolutionIO, ReferenceIO)

    def _evaluate(self, solution: SolutionBase) -> Union[np.ndarray, float]:
        """
        Calculate relative difference in area compared to reference.

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray
            The relative difference in area comared to reference.
        """
        area_ref = simpson(self.reference.solution, x=self.reference.time, axis=0)
        area_new = simpson(solution.solution, x=solution.time, axis=0)

        return abs(area_ref - area_new) / area_ref


class Shape(DifferenceBase):
    """
    Shape similarity difference metric.

    The similarity is calculated using the Pearson correlation between the reference
    and solution profiles, as well as the time offset between their peak positions, and
    the peak height of the solution profile. Additionally, if `use_derivative` is set to
    True, the similarity is also calculated using the Pearson correlation of the derivative
    profiles, and the minimum and maximum peak heights of the derivative profile.

    Attributes
    ----------
    n_metrics : int
        Number of similarity metrics calculated by the class.
    labels : list of str
        List of labels for each similarity metric calculated by the class.

    Raises
    ------
    CADETProcessError
        If `components` is not None and has more than one element.

    Notes
    -----
    Currently, this class only works for single-component systems with one peak.
    """

    _valid_references = (SolutionIO, ReferenceIO)

    @wraps(DifferenceBase.__init__)
    def __init__(
        self,
        *args: Any,
        use_derivative: bool = True,
        normalize_metrics: bool = True,
        normalization_factor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize Shape metric.

        Parameters
        ----------
        *args :
            Positional arguments for DifferenceBase.
        use_derivative : bool, optional
            If True, use the derivative profiles to calculate similarity.
            Default is True.
        normalize_metrics : bool, optional
            If True, normalize the similarity metrics to a range of [0, 1] using a
            sigmoid function.
            Default is True.
        normalization_factor : float, optional
            Normalization factor used by the sigmoid function.
            Default is None, which sets it to 1/10 of the simulation time.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, **kwargs)

        if self.reference.n_comp > 1 and not self.use_total_concentration:
            if self.components is not None and len(self.components) == 1:
                pass
            else:
                raise CADETProcessError(
                    "Shape currently only supports single component."
                )

        self.use_derivative = use_derivative
        if use_derivative:
            self.reference_der = self.reference.derivative

        self.normalize_metrics = normalize_metrics
        if normalization_factor is None:
            normalization_factor = self.reference.time[-1] / 10
        self.normalization_factor = normalization_factor

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        n_metrics = 2

        if self.use_derivative:
            n_metrics += 1

        return n_metrics

    @property
    def labels(self) -> list:
        """list[str]: List of difference metric names."""
        labels = ["Pearson Correleation", "Time offset"]
        if self.use_derivative:
            labels += ["Pearson Correlation Derivative"]
        return labels

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Evaluate the Shape similarity using Pearson correlation.

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray
            Array of similarity metrics.
        """
        corr, offset_original = pearson(
            self.reference.time,
            self.reference.solution_interpolated.solutions[0],
            solution.solution_interpolated.solutions[0],
        )

        if self.normalize_metrics:
            offset = sigmoid_distance(
                offset_original, target=0, normalization=self.normalization_factor
            )
        else:
            offset = np.abs(offset_original)

        if not self.use_derivative:
            return np.array([corr, offset])

        corr_der = pearson_offset(
            self.reference_der.time,
            self.reference_der.solution_interpolated.solutions[0],
            solution.derivative.solution_interpolated.solutions[0],
            offset_original,
        )

        return np.array(
            [
                corr,
                offset,
                corr_der,
            ]
        )


class ShapeFront(Shape):
    """
    ShapeFront similarity difference metric.

    This class extends the Shape metric by focusing on the front of a peak for
    similarity calculations.

    Attributes
    ----------
    n_metrics : int
        Number of similarity metrics calculated by the class.
    labels : list of str
        List of labels for each similarity metric calculated by the class.

    Notes
    -----
    This class is designed for single-component systems with one peak.
    """

    def __init__(
        self,
        *args: Any,
        min_percent: Optional[float] = 0.02,
        max_percent: Optional[float] = 0.98,
        use_max_slope: Optional[bool] = False,
        **kwargs: Any,
    ) -> None:
        """
        Initialize ShapeFront metric.

        Parameters
        ----------
        *args
            Positional arguments for Shape.
        min_percent : Optional[float], default=0.02
            The minimum percentage value for the metric.
        max_percent : Optional[float], default=0.98
            The maximum percentage value for the metric.
        use_max_slope : Optional[bool], default=False
            If True, use the maximum slope in calculations.
        **kwargs : dict
            Keyword arguments for Shape.
        """
        super().__init__(*args, **kwargs)

        self.min_percent = min_percent
        self.max_percent = max_percent
        self.use_max_slope = use_max_slope

        self.reference_front, idx_min, idx_max = slice_solution_front(
            self.reference,
            self.min_percent,
            self.max_percent,
            use_max_slope=use_max_slope,
            return_indices=True,
        )

        if self.use_derivative:
            self.reference_der_front = copy.deepcopy(self.reference_der)
            self.reference_der_front.solution[:] = 0
            self.reference_der_front.solution[idx_min : idx_max + 1] = (
                self.reference_der.solution[idx_min : idx_max + 1]
            )
            self.reference_der_front.update_solution()

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Evaluate the ShapeFront similarity using Pearson correlation.

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray
            Array of similarity metrics.
        """
        times = solution.time

        solution_front, idx_min, idx_max = slice_solution_front(
            solution,
            self.min_percent,
            self.max_percent,
            use_max_slope=self.use_max_slope,
            return_indices=True,
        )

        corr, offset_original = pearson(
            times,
            self.reference_front.solution_interpolated.solutions[0],
            solution_front.solution_interpolated.solutions[0],
        )

        if self.normalize_metrics:
            offset = sigmoid_distance(
                offset_original, target=0, normalization=self.normalization_factor
            )
        else:
            offset = np.abs(offset_original)

        if not self.use_derivative:
            return np.array([corr, offset])

        solution_der = solution.derivative

        solution_der_front = copy.deepcopy(solution_der)
        solution_der_front.solution[:] = 0
        solution_der_front.solution[idx_min : idx_max + 1] = (
                solution_der.solution[idx_min : idx_max + 1]
        )
        solution_der_front.update_solution()

        corr_der = pearson_offset(
            self.reference_der_front.time,
            self.reference_der_front.solution_interpolated.solutions[0],
            solution_der_front.solution_interpolated.solutions[0],
            offset_original,
        )

        return np.array(
            [
                corr,
                offset,
                corr_der,
            ]
        )


class PeakHeight(DifferenceBase):
    """
    Absolute difference in peak height difference metric.

    Attributes
    ----------
    find_minima : bool
        Indicates whether the minima instead of maxima of the peaks are found.
    reference_peaks : list of tuple
        Contains the peaks found in the reference.
    normalize_metrics : bool
        Indicates whether normalization is applied to the peak height difference scores.
    normalization_factor : list of list
        Contains the normalization factors for each peak in each component.
    """

    _valid_references = (SolutionIO, ReferenceIO)

    @wraps(DifferenceBase.__init__)
    def __init__(
        self,
        *args: Any,
        find_minima: bool = False,
        normalize_metrics: bool = True,
        normalization_factor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the PeakHeight object.

        Parameters
        ----------
        *args : tuple
            Positional arguments passed to the base class constructor.
        find_minima : bool, optional
            If True, finds the minima instead of maxima of the peaks, by default False.
        normalize_metrics : bool, optional
            If True, applies normalization to the peak height difference scores.
            The default is True.
        normalization_factor : int or None, optional
            If not None, sets the normalization factor to a constant value for all peaks.
            If None, calculates the normalization factor based on the reference peaks
            The default is None.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, **kwargs)

        self.find_minima = find_minima
        self.reference_peaks = find_peaks(self.reference, find_minima=find_minima)

        self.normalize_metrics = normalize_metrics
        if normalization_factor is None:
            normalization_factor = [
                [peak[1] for peak in self.reference_peaks[i]]
                for i in range(self.reference.n_comp)
            ]
        else:
            normalization_factor = [
                len(self.reference_peaks[i]) * [normalization_factor]
                for i in range(self.reference.n_comp)
            ]
        self.normalization_factor = normalization_factor

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        return sum([len(peak) for peak in self.reference_peaks])

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Calculate difference in peak height (concentration).

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray
            Array of difference in peak height (concentration).
        """
        solution_peaks = find_peaks(solution, find_minima=self.find_minima)

        if self.normalize_metrics:
            score = [
                sigmoid_distance(sol[1], ref[1], factor)
                for i in range(self.reference.n_comp)
                for ref, sol, factor in zip(
                    self.reference_peaks[i],
                    solution_peaks[i],
                    self.normalization_factor,
                )
            ]
        else:
            score = [
                ref[1] - sol[1]
                for i in range(self.reference.n_comp)
                for ref, sol in zip(self.reference_peaks[i], solution_peaks[i])
            ]

        return np.abs(score).flatten()


class PeakPosition(DifferenceBase):
    """
    Absolute difference in peak peak position difference metric.

    Attributes
    ----------
    reference_peaks : list of tuple
        Contains the peaks found in the reference.
    normalize_metrics : bool
        Indicates whether normalization is applied to the peak height difference scores.
    normalization_factor : list of list
        Contains the normalization factors for each peak in each component.
    """

    _valid_references = (SolutionIO, ReferenceIO)

    @wraps(DifferenceBase.__init__)
    def __init__(
        self,
        *args: Any,
        normalize_metrics: bool = True,
        normalization_factor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize PeakPosition object.

        Parameters
        ----------
        *args : tuple
            Positional arguments to pass to the parent class constructor.
        normalize_metrics : bool, optional
            Whether to normalize the difference metrics, by default True.
        normalization_factor : float or list of float, optional
            Normalization factor(s) to use for the difference metric(s). If a single
            float is given, it will be used for all the metrics. If a list of floats is
            given, it must have one element per component in the reference, by default
            None.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, **kwargs)

        self.reference_peaks = find_peaks(self.reference)

        self.normalize_metrics = normalize_metrics
        if normalization_factor is None:
            normalization_factor = [
                [peak[0] for peak in self.reference_peaks[i]]
                for i in range(self.reference.n_comp)
            ]
        else:
            normalization_factor = [
                len(self.reference_peaks[i]) * [normalization_factor]
                for i in range(self.reference.n_comp)
            ]

        self.normalization_factor = normalization_factor

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        return sum([len(peak) for peak in self.reference_peaks])

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Calculate difference in peak position (time).

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray
            The difference in peak position.
        """
        solution_peaks = find_peaks(solution)

        if self.normalize_metrics:
            score = [
                sigmoid_distance(sol[0], ref[0], factor)
                for i in range(self.reference.n_comp)
                for ref, sol, factor in zip(
                    self.reference_peaks[i],
                    solution_peaks[i],
                    self.normalization_factor[i],
                )
            ]
        else:
            score = [
                ref[0] - sol[0]
                for i in range(self.reference.n_comp)
                for ref, sol in zip(self.reference_peaks[i], solution_peaks[i])
            ]

        return np.abs(score).flatten()


class BreakthroughHeight(DifferenceBase):
    """
    Absolute difference in breakthrough curve height difference metric.

    Attributes
    ----------
    normalize_metrics : bool
        Whether to normalize the difference scores based on a sigmoid function.
    reference_bt : list of tuple
        List of breakthrough curves in the reference solution.
    """

    _valid_references = (SolutionIO, ReferenceIO)

    @wraps(DifferenceBase.__init__)
    def __init__(
        self, *args: Any, normalize_metrics: bool = True, **kwargs: Any
    ) -> None:
        """
        Initialize BreakthroughHeight metric.

        Parameters
        ----------
        *args : tuple
            Positional arguments to pass to the base class constructor.
        normalize_metrics : bool, optional
            Whether to normalize the difference scores based on a sigmoid function.
            Default is True.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, **kwargs)

        self.normalize_metrics = normalize_metrics
        self.reference_bt = find_breakthroughs(self.reference)

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        return len(self.reference_bt)

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Calculate difference in breakthrough height (concentration).

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray:
            The difference in breakthrough height.
        """
        solution_bt = find_breakthroughs(solution)

        if self.normalize_metrics:
            score = [
                sigmoid_distance(sol[1], ref[1])
                for ref, sol in zip(self.reference_bt, solution_bt)
            ]
        else:
            score = [ref[1] - sol[1] for ref, sol in zip(self.solution_bt, solution_bt)]

        return np.abs(score)


class BreakthroughPosition(DifferenceBase):
    """Absolute difference in breakthrough curve position difference metric."""

    _valid_references = (SolutionIO, ReferenceIO)

    @wraps(DifferenceBase.__init__)
    def __init__(
        self,
        *args: Any,
        normalize_metrics: bool = True,
        normalization_factor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the BreakthroughPosition object.

        Parameters
        ----------
        *args :
            Positional arguments for DifferenceBase.
        normalize_metrics : bool, optional
            Whether to normalize the metrics. Default is True.
        normalization_factor : float, optional
            Factor to use for normalization.
            If None, it is set to the maximum of the difference between the reference
            breakthrough and the start time, and the difference between the end time and
            the reference breakthrough.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, **kwargs)

        self.reference_bt = find_breakthroughs(self.reference)

        t_min = self.reference.time[0]
        t_max = self.reference.time[-1]

        self.normalize_metrics = normalize_metrics
        if normalization_factor is None:
            normalization_factor = max(
                self.reference_bt[0] - t_min, t_max - self.reference_bt[0]
            )
        self.normalization_factor = normalization_factor

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics."""
        return len(self.reference_bt)

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Calculate difference in breakthrough position (time).

        Parameters
        ----------
        solution : SolutionIO
            Concentration profile of simulation.

        Returns
        -------
        np.ndarray:
            The difference in breakthrough position.
        """
        solution_bt = find_breakthroughs(solution)

        if self.normalize_metrics:
            score = [
                sigmoid_distance(sol[0], ref[0], self.normalization_factor)
                for ref, sol in zip(self.reference_bt, solution_bt)
            ]
        else:
            score = [ref[0] - sol[0] for ref, sol in zip(self.solution_bt, solution_bt)]

        return np.abs(score)


class FractionationSSE(DifferenceBase):
    """Fractionation based score using SSE."""

    _valid_references = FractionationReference

    @wraps(DifferenceBase.__init__)
    def __init__(
        self,
        *args: Any,
        normalize_metrics: bool = True,
        normalization_factor: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the FractionationSSE object.

        Parameters
        ----------
        *args :
            Positional arguments for DifferenceBase.
        normalize_metrics : bool, optional
            Whether to normalize the metrics. Default is True.
        normalization_factor : float, optional
            Factor to use for normalization.
            If None, it is set to the maximum of the difference between the reference
            breakthrough and the start time, and the difference between the end time and
            the reference breakthrough.
        **kwargs : dict
            Keyword arguments passed to the base class constructor.
        """
        super().__init__(*args, resample=False, only_transforms_array=False, **kwargs)

        if not isinstance(self.reference, FractionationReference):
            raise TypeError(
                "FractionationSSE can only work with FractionationReference"
            )

        def transform(solution: SolutionIO) -> SolutionIO:
            """Transform solution."""
            solution = copy.deepcopy(solution)
            solution_fractions = [
                solution.create_fraction(frac.start, frac.end)
                for frac in self.reference.fractions
            ]

            solution.time = np.array(
                [(frac.start + frac.end) / 2 for frac in solution_fractions]
            )
            solution.solution = np.array(
                [frac.concentration for frac in solution_fractions]
            )

            return solution

        self.transform = transform

    def _evaluate(self, solution: SolutionIO) -> np.ndarray:
        """
        Calculate the sum of squared errors between the solution and a reference solution.

        Parameters
        ----------
        solution : SolutionIO
            An object containing the concentration profile of the simulation to be evaluated.

        Returns
        -------
        np.ndarray
            The sum of squared errors (SSE) between the solution and the reference solution.
        """
        sse = calculate_sse(solution.solution, self.reference.solution)

        return sse



import copy
import functools
import importlib
from typing import Any, Iterator, Optional

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from CADETProcess import CADETProcessError, SimulationResults, plotting
from CADETProcess.comparison import DifferenceBase
from CADETProcess.dataStructure import String, Structure, get_nested_value
from CADETProcess.numerics import round_to_significant_digits
from CADETProcess.solution import SolutionBase


class Comparator(Structure):
    """
    Class for comparing simulation results against reference data.

    Attributes
    ----------
    name : str
        Name of the Comparator instance.
    references : dict
        Dictionary containing the reference data to be compared against.
    solution_paths : dict
        Dictionary containing the solution path for each difference metric.
    metrics : list
        List of difference metrics to be evaluated.
    """

    name = String()

    def __init__(
        self,
        name: Optional[str] = None,
    ) -> None:
        """
        Initialize a new Comparator instance.

        Parameters
        ----------
        name : Optional[str]
            Name of the Comparator instance.
        """
        self.name = name
        self._metrics = []
        self.references: dict[str, SolutionBase] = {}
        self.solution_paths = {}

    def add_reference(
        self,
        reference: SolutionBase,
        update: Optional[bool] = False,
        smooth: Optional[bool] = False,
    ) -> None:
        """
        Add reference to the Comparator.

        Parameters
        ----------
        reference : SolutionBase
            Reference for comparison with SimulationResults.
        update : Optional[bool], default=False
            If True, update existing reference.
        smooth : Optional[bool], default=False
            If True, smooth data before comparison.

        Raises
        ------
        TypeError
            If reference is not an instance of SolutionBase.
        CADETProcessError
            If Reference already exists.

        """
        if not isinstance(reference, SolutionBase):
            raise TypeError("Expeced SolutionBase")

        if reference.name in self.references and not update:
            raise CADETProcessError("Reference already exists")

        reference = copy.deepcopy(reference)
        if smooth:
            reference.smooth_data()

        self.references[reference.name] = reference

    @property
    def metrics(self) -> list[DifferenceBase]:
        """list[DifferenceBase]: List of difference metrics."""
        return self._metrics

    @property
    def n_difference_metrics(self) -> int:
        """int: Number of difference metrics in the Comparator."""
        return len(self.metrics)

    @property
    def n_metrics(self) -> int:
        """int: Number of metrics to be evaluated."""
        return sum([metric.n_metrics for metric in self.metrics])

    @property
    def bad_metrics(self) -> list[float]:
        """list[float]: Worst case metrics for all difference metrics."""
        bad_metrics = [metric.bad_metrics for metric in self.metrics]

        return np.hstack(bad_metrics).flatten().tolist()

    @property
    def labels(self) -> list[str]:
        """list[str]: List of metric labels."""
        labels = []
        for metric in self.metrics:
            try:
                metric_labels = metric.labels
            except AttributeError:
                metric_labels = [f"{metric}"]
                if metric.n_metrics > 1:
                    metric_labels = [f"{metric}_{i}" for i in range(metric.n_metrics)]

            if len(metric_labels) != metric.n_metrics:
                raise CADETProcessError(f"Must return {metric.n_labels} labels.")

            labels += metric_labels

        return labels

    @functools.wraps(DifferenceBase.__init__)
    def add_difference_metric(
        self,
        difference_metric: str,
        reference: str | SolutionBase,
        solution_path: str,
        *args: Any,
        **kwargs: Any,
    ) -> DifferenceBase:
        """
        Add a difference metric to the Comparator.

        Parameters
        ----------
        difference_metric : str
            Name of the difference metric to be evaluated.
        reference : str | SolutionBase
            Name of the reference or reference itself.
        solution_path : str
            Path to the solution in SimulationResults.
        *args, **kwargs
            Additional arguments and keyword arguments to be passed to the
            difference metric constructor.

        Returns
        -------
        DifferenceBase
            The new difference metric instance.

        Raises
        ------
        CADETProcessError
            If the difference metric or reference is unknown.
        """
        try:
            module = importlib.import_module("CADETProcess.comparison.difference")
            cls_ = getattr(module, difference_metric)
        except KeyError:
            raise CADETProcessError("Unknown Metric Type.")

        if isinstance(reference, SolutionBase):
            reference = reference.name

        if reference not in self.references:
            raise CADETProcessError("Unknown Reference.")

        reference = self.references[reference]

        metric = cls_(reference, *args, **kwargs)

        self.solution_paths[metric] = solution_path

        self._metrics.append(metric)

        return metric

    def extract_solution(
        self,
        simulation_results: SimulationResults,
        metric: DifferenceBase,
    ) -> SolutionBase:
        """
        Extract the solution for a given metric from the SimulationResults object.

        Parameters
        ----------
        simulation_results : SimulationResults
            The SimulationResults object containing the solution.
        metric : DifferenceBase
            The difference metric object for which to extract the solution.

        Returns
        -------
        SolutionBase
            The solution for the given metric.

        Raises
        ------
        CADETProcessError
            If the solution path for the given metric is not found.
        """
        try:
            solution_path = self.solution_paths[metric]
            solution = get_nested_value(simulation_results.solution_cycles, solution_path)[-1]
        except KeyError:
            raise CADETProcessError("Could not find solution path")

        return copy.deepcopy(solution)

    def evaluate(self, simulation_results: SimulationResults) -> list[float]:
        """
        Evaluate all metrics for a given simulation and return the results as a list.

        Parameters
        ----------
        simulation_results : SimulationResults
            The SimulationResults object containing the solutions for all metrics.

        Returns
        -------
        list[float]
            A list containing the values of all difference of metrics after comparison.
        """
        metrics = []
        for metric in self.metrics:
            solution = self.extract_solution(simulation_results, metric)
            m = metric.evaluate(solution)
            metrics.append(m)

        metrics = np.hstack(metrics).tolist()

        return metrics

    __call__ = evaluate

    def setup_comparison_figure(
        self,
        plot_individual: Optional[bool] = False,
    ) -> tuple[list[Figure], npt.NDArray[Axes]]:
        """
        Set up a figure for comparing simulation results.

        Parameters
        ----------
        plot_individual : Optional[bool], default=False
            If True, return figures for individual metrics.
            Otherwise, return a single figure for all metrics.

        Returns
        -------
        tuple
            A tuple containing:
            - list[plt.Figure]: A list of Matplotlib Figure objects.
            - npt.NDArray[plt.Axes]: An array of Axes objects with one Axes per
              difference metric.
        """
        if self.n_difference_metrics == 0:
            return (None, None)

        comparison_fig_all, comparison_axs_all = plotting.setup_figure(
            n_rows=self.n_difference_metrics, squeeze=False
        )

        plt.close(comparison_fig_all)
        comparison_axs_all = comparison_axs_all.reshape(-1)

        comparison_fig_ind: list[Figure] = []
        comparison_axs_ind: list[Axes] = []
        for i in range(self.n_difference_metrics):
            fig, axs = plt.subplots()
            comparison_fig_ind.append(fig)
            comparison_axs_ind.append(axs)
            plt.close(fig)

        comparison_axs_ind = np.array(comparison_axs_ind).reshape(comparison_axs_all.shape)

        if plot_individual:
            return comparison_fig_ind, comparison_axs_ind
        else:
            return comparison_fig_all, comparison_axs_all

    def plot_comparison(
        self,
        simulation_results: SimulationResults,
        axs: Optional[Axes | list[Axes]] = None,
        figs: Optional[Figure | list[Figure]] = None,
        file_name: Optional[str] = None,
        show: Optional[bool] = True,
        plot_individual: Optional[bool] = False,
        x_axis_in_minutes: Optional[bool] = True,
    ) -> tuple[list[Figure], npt.NDArray[plt.Axes]]:
        """
        Plot the comparison of the simulation results with the reference data.

        Parameters
        ----------
        simulation_results : SimulationResults
            Simulation results to compare to reference data.
        axs : Optional[Axes | list[Axes]], default=None
            An array of Axes objects to use for plotting the metrics.
        figs : Optional[Figure | list[Figure]]
            List of figures to use for plotting the metrics.
        file_name : Optional[str]
            Name of the file to save the figure to.
        show : Optional[bool], default=True
            If True, displays the figure(s) on the screen.
        plot_individual : Optional[bool], default=False
            If True, generates a separate figure for each metric.
        x_axis_in_minutes: Optional[bool], default=True
            If True, the x-axis will be plotted using minutes. The default is True.

        Returns
        -------
        tuple
            A tuple containing:
            - list[plt.Figure]: A list of Matplotlib Figure objects.
            - npt.NDArray[plt.Axes]: An array of Axes objects with one Axes per
              difference metric.
        """
        if axs is None:
            figs, axs = self.setup_comparison_figure(plot_individual)
        if not isinstance(figs, list):
            figs = [figs]

        for ax, metric in zip(axs, self.metrics):
            solution = self.extract_solution(simulation_results, metric)
            solution_sliced = metric.slice_and_transform(solution)

            y_max = 1.1 * max(np.max(solution_sliced.solution), np.max(metric.reference.solution))

            fig, ax = solution_sliced.plot(
                ax=ax,
                show=False,
                y_max=y_max,
            )

            plot_args = {
                "linestyle": "dotted",
                "color": "k",
                "label": "reference",
            }
            ref_time = metric.reference.time
            if x_axis_in_minutes:
                ref_time = ref_time / 60

            plotting.add_overlay(ax, metric.reference.solution, ref_time, **plot_args)
            ax.legend(loc=1)

            m = metric.evaluate(solution_sliced, slice=False)
            m = round_to_significant_digits(m, digits=2)

            text = f"{metric}: "
            if metric.n_metrics > 1:
                try:
                    text += "\n"
                    for i, (label, m) in enumerate(zip(metric.labels, m)):
                        text += f"{label}: ${m}$"
                        if i < metric.n_metrics - 1:
                            text += " \n"
                except AttributeError:
                    text += f"{m}"
            else:
                text += str(m[0])

            plotting.add_text(ax, text, fontsize=14)

        for fig in figs:
            fig.tight_layout()
            if not show:
                plt.close(fig)
            else:
                dummy = plt.figure(figsize=fig.get_size_inches())
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                plt.show()

        if file_name is not None:
            if plot_individual:
                name, suffix = file_name.split(".")
                for fig, metric in zip(figs, self.metrics):
                    fig.savefig(f"{name}_{metric}.{suffix}")
            else:
                figs[0].savefig(file_name)

        return figs, axs

    def __iter__(self) -> Iterator[list[DifferenceBase]]:
        """Yield metrics from the instance."""
        yield from self.metrics

    def __str__(self) -> str:
        """str: Name of the Comparator."""
        if self.name is not None:
            return self.name
        else:
            return self.__class__.__name__



import unittest

from CADETProcess.optimization import ResultsCache


class TestCache(unittest.TestCase):
    def setUp(self):
        self.cache_dict = ResultsCache(use_diskcache=False)
        self.cache_disk = ResultsCache(use_diskcache=True)

        key = (None, "objective", str([1, 2, 3]))
        result = "important result"
        self.cache_dict.set(key, result)
        self.cache_disk.set(key, result)

        key = (None, "intermediate", str([1, 2, 3]))
        result = "temporary result"
        self.cache_dict.set(key, result, "temp")
        self.cache_disk.set(key, result, "temp")

    def tearDown(self):
        self.cache_disk.delete_database()

    def test_set(self):
        new_result = "new"
        key = (None, "other", str([1, 2, 3]))

        self.cache_dict.set(key, new_result)
        cached_result = self.cache_dict.get(key)
        self.assertEqual(cached_result, new_result)

        self.cache_disk.set(key, new_result)
        cached_result = self.cache_disk.get(key)
        self.assertEqual(cached_result, new_result)

    def test_get(self):
        key = (None, "intermediate", str([1, 2, 3]))
        result_expected = "temporary result"
        cached_result = self.cache_dict.get(key)
        self.assertEqual(result_expected, cached_result)

        key = (None, "intermediate", str([1, 2, 3]))
        result_expected = "temporary result"
        cached_result = self.cache_disk.get(key)
        self.assertEqual(result_expected, cached_result)

        key = (None, "false", str([1, 2, 3]))
        with self.assertRaises(KeyError):
            cached_result = self.cache_dict.get(key)

        with self.assertRaises(KeyError):
            cached_result = self.cache_disk.get(key)

    def test_delete(self):
        key = (None, "intermediate", str([1, 2, 3]))
        self.cache_dict.delete(key)
        with self.assertRaises(KeyError):
            cached_result = self.cache_dict.get(key)

        self.cache_disk.delete(key)
        with self.assertRaises(KeyError):
            cached_result = self.cache_disk.get(key)

    def test_tags(self):
        tags_expected = ["temp"]
        tags = list(self.cache_dict.tags.keys())
        self.assertEqual(tags_expected, tags)

        key = (None, "other", str([1, 2, 4]))
        self.cache_dict.set(key, "new", "foo")
        tags_expected = ["temp", "foo"]
        tags = list(self.cache_dict.tags.keys())
        self.assertEqual(tags_expected, tags)

    def test_prune(self):
        key = (None, "intermediate", str([1, 2, 3]))
        self.cache_dict.prune("temp")
        with self.assertRaises(KeyError):
            cached_result = self.cache_dict.get(key)

        key = (None, "objective", str([1, 2, 3]))
        result_expected = "important result"
        cached_result = self.cache_dict.get(key)
        self.assertEqual(result_expected, cached_result)

        key = (None, "intermediate", str([1, 2, 3]))
        self.cache_disk.prune("temp")
        with self.assertRaises(KeyError):
            cached_result = self.cache_disk.get(key)

        key = (None, "objective", str([1, 2, 3]))
        result_expected = "important result"
        cached_result = self.cache_disk.get(key)
        self.assertEqual(result_expected, cached_result)

    def test_delete_database(self):
        pass


if __name__ == "__main__":
    unittest.main()
